"""Set of functions to import, export or simply convert bounding box numpy arrays
or dataframe depending on the convention used.

12 compatible formats, compatible with

- cAIpy JSON
- COCO
- darknet
- etc...
"""

from collections.abc import Sequence
from copy import deepcopy

import numpy as np
import pandas as pd
from numpy import ndarray

OPTIONS = {
    "cxwcyh": {"center": True, "size": True, "point": False, "order": [0, 1, 2, 3]},
    "xwyh": {"center": False, "size": True, "point": False, "order": [0, 1, 2, 3]},
    "xxyy": {"center": False, "size": False, "point": False, "order": [0, 1, 2, 3]},
    "xy": {"center": True, "size": False, "point": True, "order": [0, 1]},
}

OPTIONS["cxcywh"] = {**OPTIONS["cxwcyh"], "order": [0, 2, 1, 3]}
OPTIONS["xywh"] = {**OPTIONS["xwyh"], "order": [0, 2, 1, 3]}
OPTIONS["xyxy"] = {**OPTIONS["xxyy"], "order": [0, 2, 1, 3]}

FORMAT_TO_OPTIONS = {}
for name, options in OPTIONS.items():
    FORMAT_TO_OPTIONS[name] = {**options, "relative": True}
    FORMAT_TO_OPTIONS[name.upper()] = {**options, "relative": False}


def parse_format_string(format_string: str) -> tuple[dict[str, bool], list[int]]:
    """Function used to parse the format string and convert it to a dictionary
    with options as boolean and an order list

    output dictionary keys are:

    - "center": if the first x or y coordinate is the center of the box or the minimum
      value (left for x, top for y)
    - "size": if the second coordinate is the size of the box
      (width for x, height for y), or the maximum value (right for x, bottom for y)
    - "relative": if the coordinated are in pixels of normalized with image size
      (image width for x, image height for y)

    order output list is order in which x1, x2, y1, y2 needs to be taken for the input
    array of bounding boxes. Possible values are ``[0, 1, 2, 3]`` (for e.g. ``xxyy``)
    or ``[0, 2, 1, 3]`` (for e.g. ``xyxy``)

    Args:
        format_string: format string describing the convention used for box coordinates
            can be e.g. ``XYWH`` (cAIpy / COCO format), ``cxwcyh`` (darknet format) etc.
            See documentation in :func:`.import_bbox` and :func:`.export_bbox` for a
            breakdown of the syntax.

    Raises:
        ValueError: raised when the formatstring is not in the 12 allowed values.

    Returns:
        tuple containing two elements
        - dictionary of options (see detailed keys above)
        - list of column positions for x1, x2, y1, y1 in the bbox array
    """
    valid_tokens = list(FORMAT_TO_OPTIONS)
    if format_string not in valid_tokens:
        raise ValueError(
            "invalid format string for bbox. "
            f"Valid formats are {valid_tokens}, got {format_string}"
        )
    parsed_options = deepcopy(FORMAT_TO_OPTIONS[format_string])
    order = parsed_options.pop("order")

    return parsed_options, order


def column_names_from_options(
    relative: bool, center: bool, size: bool, point: bool
) -> list[str]:
    """Generate the column names from dictionary of options, generated by
    :func:`.parse_format_string`

    This is useful to know what column to take from input dataframe when importing
    bounding boxes, or what name must be given to output dataframe when exporting
    bounding boxes

    Args:
        relative: Whether the box coordinates are relative to image size (coordinates
            are usually between 0 and 1) or not (coordinates are in pixels).
        center: Whether the first xy coordinate are linked to center of box or bottom
            left corner (in this case, it's xmin and ymin)
        size: Whether the second xy coordinates are box width and height or coordinate
            of top right corner (in this case, it's xmax and ymax). Note that size
            cannot be False if center is True.
        point: Whether the second xy coordinates are expected to be 0 (because it's only
            a point) or not.

    Returns:
        list of column names in the right order
    """
    if relative:
        suffix = "_relative"
    else:
        suffix = ""

    if point:
        column_names = ["box_x_min", "box_y_min"]
    elif center:
        column_names = ["box_x_center", "box_width", "box_y_center", "box_height"]
    elif size:
        column_names = ["box_x_min", "box_width", "box_y_min", "box_height"]
    else:
        column_names = ["box_x_min", "box_x_max", "box_y_min", "box_y_max"]

    column_names = [f"{c}{suffix}" for c in column_names]
    return column_names


def column_names_from_format_string(format_string: str) -> list[str]:
    """Generate the column names from the format string.

    Essentially, it calls a composition of :func:`.parse_format_string` and
    :func:`.column_names_from_options`

    This function is useful when you want to drop these columns from the annotations
    dataframe in you Dataset object

    Args:
        format_string: short acronym describing the box format

    Returns:
        list[str]: List of verbose column names associated to given format.
    """
    format_options, order = parse_format_string(format_string)
    column_names = column_names_from_options(**format_options)
    return [column_names[i] for i in order]


def import_bbox(
    bounding_boxes: pd.DataFrame | ndarray,
    images_df: pd.DataFrame,
    image_ids: Sequence[int] | pd.Series | np.ndarray | None = None,
    input_format: str = "XYWH",
) -> pd.DataFrame:
    """Convert bounding boxes from a particular format to cAIpy/COCO reference.
    Essentially, this will convert bounding box coordinates to pixel coordinates with
    box size.

    ``input_format`` is a format string following a particular syntax to quickly
    let the function figure what conversion operation to do.

    - ``cx`` stands for x-coordinate of the center of the box
    - first ``x`` stands for x-coordinate of the leftmost point of the box
      (i.e. the minimum x value)
    - second ``x`` stands for x-coordinate of the rightmost point of the box
      (i.e. the maximum x value)
    - ``w`` stands for the box width
    - ``cy`` stands for the y-coordinates of the center of the box
    - first ``y`` stands for y-coordinate of the upper point of the box
      (i.e. the minimum y value)
    - second ``y`` stands for y-coordinate of the bottom point of the box
      (i.e. the maximum y value)
    - ``h`` stands fo the box height

    In addition, letters in uppercase indicate that the coordinate are in pixels,
    while lowercase indicate that they are relative to the image size, i.e. 1 is
    the full width of height of the image.

    Example:

        cAIpy / COCO::

            "XYWH"  # x left, y top, width, height, all in pixels

        darknet::

            "cxwcyh"  # x center, width, y center, height, all in relative

    Note:

        - x-coordinates and y-coordinates must follow the same convention.
          As such, examples like ``xxYH`` are not allowed
        - order can be either ``{x1}{x2}{y1}{y2}`` or ``{x1}{y1}{x2}{y2}``.
          If ``bounding_boxes`` is a dataframe, it should not change anything as
          columns are named, but order is important if it is a numpy array instead
        - center coordinate can only be used with box size.
          ``cxxcyy`` is thus not allowed
        - Finally, only 12 options are allowed options. Here is the list :
          ``cxwcyh``, ``xwyh``, ``xxyy``, ``cxcywh``,
          ``xywh``, ``xyxy``, ``CXWCYH``, ``XWYH``, ``XXYY``, ``CXCYWH``,
          ``XYWH``, ``XYXY``

    In the case ``bounding_boxes`` is a dataframe, columns must correspond to the
    given format string. The convention is the following:

    - center coordinates are named ``box_{x/y}_center``
    - extremal coordinated are named ``box_{x/y}_{min/max}``
    - box sizes are named ``box_{width/height}``
    - If coordinates are relative, names get appended the ``_relative`` suffix

    Example:

        cAipy / COCO::

            "XYWH" # ["box_x_min", "box_y_min", "box_width", "box_height"]

        darknet::

            "cxwcyh"
            # ["box_x_center_relative", "box_width_relative",
            #  "box_y_center_relative", "box_height_relative"]

    Args:
        bounding_boxes: array or dataframe containing the bbox coordinates

            - If it's an array, ``image_ids`` must be given, and its shape must be (4, N)
            - If it's a dataframe, corresponding columns of ``input_format`` must be
              present (see example above)
        images_df: dataframe containing image data, and especially image sizes.
        image_ids: optional list to link every bounding box to its corresponding image
            with its id. As such, this must be the same length as ``bounding_boxes``
            array. Note that if ``bounding_boxes`` is an array and has a ``image_id``
            column, this option is not used. Defaults to None.
        input_format: string defining the expected input format. Defaults to "XYWH".

    Raises:
        ValueError: This error is raised if the input format string is not well
            formatted or if bounding boxes and image ids shapes are incompatible

    Returns:
        DataFrame containing the reference cAIpy format, i.e. with x left, width, y top
        and height, in pixels. index is the same one os ``bounding_boxes``, and columns
        are, as described above, ``box_x_min``, ``box_width``, ``box_y_min`` and
        ``box_height``. This is the format expected in the
        :class:`Dataset <lours.dataset.Dataset>` or
        :class:`Evaluator <lours.evaluation.Evaluator`
    """
    input_format_options, order = parse_format_string(input_format)
    if isinstance(bounding_boxes, ndarray):
        n_col = bounding_boxes.shape[1]
        if n_col != 4:
            raise ValueError(
                "Error with input bounding_boxes array shape. "
                f"Expected (N, 4), got {bounding_boxes.shape} instead"
            )
        bbox_array = bounding_boxes[:, order]
        index = None
    else:
        index = bounding_boxes.index
        bbox_array = bounding_boxes[
            column_names_from_options(**input_format_options)
        ].to_numpy()
        if "image_id" in bounding_boxes and image_ids is None:
            image_ids = bounding_boxes["image_id"]

    if image_ids is None:
        raise ValueError(
            "Expected image_ids to be either given in bounding_boxes "
            "dataframe within the 'image_id' column or in image_ids option"
        )

    if bbox_array.shape[0] != len(image_ids):
        raise ValueError("Expected as many image_ids as bbox_array rows")

    if input_format_options["point"]:
        if bbox_array.shape[1] != 2:
            raise ValueError("Expected 2 values per row in input array for points")
        x1, y1 = bbox_array.T
        x2, y2 = np.zeros_like(x1), np.zeros_like(y1)
    else:
        if bbox_array.shape[1] != 4:
            raise ValueError(
                "Expected 4 values per row in input array for bounding boxes"
            )
        x1, x2, y1, y2 = bbox_array.T

    if input_format_options["relative"]:
        im_width = (
            images_df.loc[image_ids, "width"]  # pyright: ignore
            .to_numpy()
            .astype(float)
        )
        im_height = (
            images_df.loc[image_ids, "height"]  # pyright: ignore
            .to_numpy()
            .astype(float)
        )
        x1 *= im_width
        y1 *= im_height

        if not input_format_options["point"]:
            x2 *= im_width
            y2 *= im_height

    if input_format_options["center"]:
        xmin, width = x1 - x2 / 2, x2
        ymin, height = y1 - y2 / 2, y2
    elif input_format_options["size"]:
        xmin, width, ymin, height = x1, x2, y1, y2
    else:
        xmin, width = x1, x2 - x1
        ymin, height = y1, y2 - y1
    converted_bbox_array = np.stack([xmin, ymin, width, height], axis=1)
    bbox_df = pd.DataFrame(
        converted_bbox_array,
        columns=column_names_from_format_string("XYWH"),
    )
    if index is not None:
        bbox_df.index = index
    return bbox_df


def export_bbox(
    annotations_df: pd.DataFrame,
    images_df: pd.DataFrame,
    input_format: str = "XYWH",
    output_format: str = "XYWH",
) -> pd.DataFrame:
    """Convert bounding boxes in Lours's reference to a desired format.

    ``output_format`` is a format string following the same syntax as ``input_format``
    in :func:`.import_bbox`.

    - ``cx`` stands for x-coordinate of the center of the box
    - first ``x`` stands for x-coordinate of the leftmost point of the box
      (i.e. the minimum x value)
    - second ``x`` stands for x-coordinate of the rightmost point of the box
      (i.e. the maximum x value)
    - ``w`` stands for the box width
    - ``cy`` stands for the y-coordinates of the center of the box
    - first ``y`` stands for y-coordinate of the upper point of the box
      (i.e. the minimum y value)
    - second ``y`` stands for y-coordinate of the bottom point of the box
      (i.e. the maximum y value)
    - ``h`` stands fo the box height

    In addition, letters in uppercase indicate that the coordinate are in pixels,
    while lowercase indicate that they are relative to the image size, i.e. 1 is
    the full width of height of the image.

    Example:

        cAIpy / COCO::

            "XYWH"  # x left, y top, width, height, all in pixels

        darknet::

            "cxwcyh"  # x center, width, y center, height, all in relative

    Note:

        - x-coordinates and y-coordinates must follow the same convention.
          As such, examples like ``xxYH`` are not allowed
        - order can be either ``{x1}{x2}{y1}{y2}`` or ``{x1}{y1}{x2}{y2}``.
          Order is important if the output dataframe is expected to be converted to
          numpy after calling this function
        - center coordinate can only be used with box size.
          ``cxxcyy`` is thus not allowed
        - Finally, only 12 options are allowed options. Here is the list :
          ``cxwcyh``, ``xwyh``, ``xxyy``, ``cxcywh``,
          ``xywh``, ``xyxy``, ``CXWCYH``, ``XWYH``, ``XXYY``, ``CXCYWH``,
          ``XYWH``, ``XYXY``

    The created dataframe will have the following column names

    - center coordinates are named ``box_{x/y}_center``
    - extremal coordinated are named ``box_{x/y}_{min/max}``
    - box sizes are named ``box_{width/height}``
    - If coordinates are relative, names get appended the ``_relative`` suffix

    Example:

        cAipy / COCO::

            "XYWH" # ["box_x_min", "box_y_min", "box_width", "box_height"]

        darknet::

            "cxwcyh"
            # ["box_x_center_relative", "box_width_relative",
            #  "box_y_center_relative", "box_height_relative"]

    Args:
        annotations_df: dataframe containing the annotations in Lours format.
            Must have at least ``box_x_min``, ``box_width``, ``box_y_min``,
            ``box_height`` and ``image_id`` columns.
        images_df: dataframe containing the image sizes in Lours format.
            Must have at least ``width``, ``height``, and the ids must correspond in the
            column ``image_id`` the annotations dataframe.
        input_format: string defining the expected input format. Defaults to "XYWH".
        output_format: string describing the desired bounding box format
            (see above). Defaults to "XWYH".

    Returns:
        dataframe containing the bounding box coordinates with the corresponding column
        names. It will have the same index as ``annotations_df``
    """
    output_format_options, order = parse_format_string(output_format)
    input_format_options, _ = parse_format_string(input_format)
    if input_format != "XYWH":
        annotations_df = import_bbox(
            annotations_df, images_df, input_format=input_format
        )
    if input_format_options["point"]:
        xmin, ymin = annotations_df[column_names_from_format_string("XY")].to_numpy().T
        width, height = xmin * 0, ymin * 0
    else:
        xmin, ymin, width, height = (
            annotations_df[column_names_from_format_string("XYWH")].to_numpy().T
        ).astype(float)
    if output_format_options["relative"]:
        im_width = images_df.loc[annotations_df["image_id"], "width"].to_numpy()
        im_height = images_df.loc[annotations_df["image_id"], "height"].to_numpy()
        xmin /= im_width
        width /= im_width
        ymin /= im_height
        height /= im_height

    if output_format_options["center"]:
        x_center = xmin + width / 2
        y_center = ymin + height / 2
        if output_format_options["point"]:
            bbox_array = np.stack([x_center, y_center], axis=1)
        else:
            bbox_array = np.stack([x_center, width, y_center, height], axis=1)
    elif output_format_options["size"]:
        bbox_array = np.stack([xmin, width, ymin, height], axis=1)
    else:
        xmax = xmin + width
        ymax = ymin + height
        bbox_array = np.stack([xmin, xmax, ymin, ymax], axis=1)
    bbox_df = pd.DataFrame(
        bbox_array,
        columns=column_names_from_options(**output_format_options),
        index=annotations_df.index,
    )
    new_column_order = [
        # order needs to be reversed to construct the dataframe the right way
        bbox_df.columns[order.index(i)]
        for i in range(len(order))
    ]
    return bbox_df[new_column_order]


def convert_bbox(
    bounding_boxes: ndarray | pd.DataFrame,
    images_df: pd.DataFrame,
    image_ids: Sequence[int] | None = None,
    input_format: str = "XYWH",
    output_format: str = "cxwcyh",
) -> pd.DataFrame:
    """Convert bounding box from a particular format to another, using a composition of
    :func:`.import_bbox` and :func:`.export_bbox`.

    See their documentation for an explanation of format strings

    Args:
        bounding_boxes: array or dataframe containing the bbox coordinates.

            - If it's an array, ``image_ids`` must be given.
            - If it's a dataframe, corresponding columns of ``input_format`` must be
              present
        images_df: dataframe containing image data, and especially image sizes.
        image_ids: optional list to link every bounding box to its corresponding image
            with its id. As such, this must be the same length as ``bounding_boxes``
            array. Note that if ``bounding_boxes`` is a dataframe and has a ``image_id``
            column, this option is not used. Defaults to None.
        input_format: string defining the expected input format. Defaults to "XYWH".
        output_format: string describing the desired bounding box format.
            Defaults to "XYWH".

    Raises:
        ValueError: This error is raised if the input format string is not well
            formatted or if bounding boxes and image ids shapes are incompatible

    Returns:
        dataframe containing the bounding box coordinates with the corresponding column
        names. It will have the same index as ``bounding_boxes`` if it is a dataframe.
    """
    imported_bbox = import_bbox(bounding_boxes, images_df, image_ids, input_format)
    if (
        isinstance(bounding_boxes, pd.DataFrame)
        and "image_id" in bounding_boxes.columns
    ):
        imported_bbox["image_id"] = bounding_boxes["image_id"]
    else:
        imported_bbox["image_id"] = image_ids
    return export_bbox(
        imported_bbox,
        images_df,
        "XYWH",
        output_format,
    )



from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Literal, cast

import pandas as pd
from tqdm.auto import tqdm

from . import bbox_converter, try_import_fiftyone

if TYPE_CHECKING:
    import fiftyone as fo
else:
    fo = try_import_fiftyone()


def make_fiftyone_compatible(
    input_df: pd.DataFrame,
    column_names: Sequence[str] = (),
    replacement_string: str = "->",
) -> tuple[pd.DataFrame, list[str]]:
    """Make column names compatible with fiftyone.

    Fiftyone is incompatible with names with a ".", so replace them with a proper
    character.

    Fiftyone is also incompatible with names starting with 'attributes', which is the
    case for libia.Dataset.annotations attributes columns, so we replace the string
    'attributes' with 'attr' in each column name.

    Note:
        If no name in ``column_names`` has a forbidden character, this function simply
        return its inputs.

    Args:
        input_df: DataFrame for which column names will be replaced
        column_names: Column names to rename. If the names are not present in the
            ``input``, no error will be raised. Defaults to ().
        replacement_string: string used to replace forbidden characters.
            Defaults to ``->``.

    Returns:
        tuple with 2 elements
         - DataFrame with modified column names from ``input``
         - List of modified names.

    """
    if not column_names:
        return input_df, []
    new_column_names = [name.replace(".", replacement_string) for name in column_names]
    # replace column names containing 'attributes' with 'attr' otherwise the fiftyone
    # web app will crash if we try to filter on these columns
    # e.g. "attributes->out_of_frame" will become "attr->out_of_frame" in fiftyone
    new_column_names = [name.replace("attributes", "attr") for name in new_column_names]
    output = input_df.rename(dict(zip(column_names, new_column_names)), axis=1)
    return output, new_column_names


def annotations_to_fiftyone(
    annotations_frame: pd.DataFrame,
    attribute_columns: Sequence[str] = (),
    bbox_column: str = "bbox",
    allow_keypoints: bool = True,
) -> pd.DataFrame:
    """Convert annotations frame into a DataFrame using the same index with
    :class:`~fiftyone.core.labels.Detection` or
    :class:`~fiftyone.core.labels.Keypoint` object in the "fo_detection" column and
    other additional fiftyone columns.

    Args:
        annotations_frame: DataFrame containing information about annotation. Most
            likely coming out of a :class:`.Dataset` object.
        attribute_columns: Columns describing real attributes. these attributes are
            differentiated from functional metadata such as ``image_id``.
            Defaults to ().
        bbox_column: column containing bounding coordinates lists. Must be already
            converted to fiftyone's xywh format. Defaults to "bbox".
        allow_keypoints: If set to True, will deduce keypoints from bounding box of size
            0. If not, every bounding box will be a detection. Defaults to True.

    Returns:
        DataFrame with the same index as ``annotations_frame``, containing
        ``fo_detection``, ``fo_id`` and ``is_keypoint`` columns.

        - ``fo_detection`` is the fiftyone object describing the detection, to be added
          to the related image sample. It can be either a
          :class:`~fiftyone.core.labels.Detection`
          or :class:`~fiftyone.core.labels.Keypoint` object
        - ``fo_id`` if the UUID given by fiftyone to identify the detection in the
          database.
        - ``is_keypoint`` is the boolean value indicating if the fiftyone object in the
          former column is a :class:`~fiftyone.core.labels.Detection` or
          :class:`~fiftyone.core.labels.Keypoint` object.
          This will make filtering much faster than looking up the type for each row.
    """

    def to_detection(row: pd.Series) -> fo.Detection:
        return fo.Detection(
            label=row["category_str"],
            label_id=row["category_id"],
            lours_id=row.name,
            bounding_box=row[bbox_column],
            **row[attribute_columns].dropna(),  # pyright: ignore
        )

    def to_keypoint(row: pd.Series) -> fo.Keypoint:
        return fo.Keypoint(
            label=row["category_str"],
            label_id=row["category_id"],
            lours_id=row.name,
            points=[row[bbox_column][:2]],
            **row[attribute_columns].dropna(),  # pyright: ignore
        )

    if allow_keypoints:
        is_keypoint = (annotations_frame["box_width"] == 0) & (
            annotations_frame["box_height"] == 0
        )
        if (~is_keypoint).sum() > 0:
            detections = annotations_frame[~is_keypoint].apply(
                to_detection,  # pyright: ignore
                axis=1,
            )
        else:
            detections = pd.Series([])
        keypoints = annotations_frame[is_keypoint].apply(
            to_keypoint,  # pyright: ignore
            axis=1,
        )
        detections = pd.DataFrame(
            {
                "is_keypoint": is_keypoint,
                "fo_detection": pd.concat([detections, keypoints]),
            }
        )
    else:
        detections = (
            annotations_frame.apply(to_detection, axis=1)  # pyright: ignore
            .rename("fo_detection")
            .to_frame()
        )
        detections["is_keypoint"] = False
    detections["fo_id"] = detections["fo_detection"].apply(lambda x: x.id)
    return detections


def create_fo_dataset(
    name: str,
    images_root: Path,
    images: pd.DataFrame,
    annotations: dict[str, pd.DataFrame],
    bounding_box_formats: dict[str, str] | None = None,
    label_map: dict[int, str] | None = None,
    image_tag_columns: Sequence[str] = (),
    annotations_attributes_columns: Sequence[str] | dict[str, Sequence[str]] = (),
    allow_keypoints: bool = False,
    existing: Literal["error", "update", "erase"] = "error",
) -> tuple[fo.Dataset, pd.Series, dict[str, pd.DataFrame]]:
    """Generic function to create a fiftyone dataset from images and annotations
    DataFrames. See :func:`.dataset_to_fiftyone` and :func:`.evaluator_to_fityone`
    for more specific functions

    Args:
        name: Name of the fiftyone dataset to add the samples to. If the dataset
            does not exist, it will be created.
        images_root: root folder for images. Fiftyone will try to load the images by
            concatenating this path and the value in images' ``relative_path`` column.
        images: DataFrame comprising image information.
            Must have at least ``relative_path`` column, and the column specified in
            ``image_tag_columns``
        annotations: dictionary of DataFrames comprising detections annotations
            information. Each entry must have at least ``image_id``, `category_id``,
            ``category_str`` (if ``label_map`` argument is None) columns,
            and the compatible columns for bbox conversion given the corresponding
            input format (see :func:`.convert_bbox`)
        bounding_box_formats: dictionary of format strings to convert bounding boxes
            of given annotations.
            For each dictionary entry in ``annotations`` dictionary, if its key is
            not included in this dictionary, the format "XYWH" (COCO cAIpy) will be
            assumed. If set to None, the format "XYWH" will always be assumed.
            Defaults to None
        label_map: dictionary comprising the category id -> category string
            correspondence, similar to :class:`.Dataset` and :class:`.Evaluator`
            label maps. If given, will populate the ``category_str`` of each annotation
            DataFrame. If set to None, will assume the column is already present.
            Defaults to None.
        image_tag_columns: List of column names to use for sample attributes in given
            ``images`` DataFrame when creating fiftyone samples. Defaults to ().
        annotations_attributes_columns: Either List of column names or dictionary of
            lists of column names. Is used to give attributes to detection which are
            then added to the created. If it's a dictionary, each annotation set in the
            ``annotations`` dictionary gets its own list of columns to use as
            attributes. If not, the ame list of column will be used for all annotations.
            Defaults to ().
        allow_keypoints: if set to True, will convert bounding boxes of size 0, 0 to
            keypoints
        existing: What to do in case there is already a fiftyone dataset with the
            same name.

            - "error": will raise an error.
            - "erase": will erase the existing dataset before uploading
                this one
            - "update": will try to update the dataset by fusing together samples
                with the same "relative_path"

            Defaults to "error".

    Returns:
        tuple with three elements
         - Fiftyone dataset that can then be used to launch the webapp with
           :func:`fiftyone.launch_app`
         - Series with the same index as ``images`` input dataframe, containing the
           fiftyone index of each image's corresponding sample. Useful when the image
           needs to be modified.
         - Dictionary of Dataframes, with the same keys as ``annotations`` input
           dictionary. Each value is a DataFrame with the same index as its
           corresponding value in ``annotations``. Its columns are ``fo_id`` and
           ``is_keypoint``

           - ``fo_id`` is the fiftyone index of each annotation (whether it is a
             :class:`~fiftyone.core.labels.Keypoint` or a
             :class:`~fiftyone.core.labels.Detection`)
           - ``is_keypoint`` is a bool column indicating if the annotation is a
             :class:`~fiftyone.core.labels.Keypoint` object or simply a
             :class:`~fiftyone.core.labels.Detection` object.
    """
    do_update = False
    if fo.dataset_exists(name):
        if existing == "update":
            result = fo.load_dataset(name)
            do_update = True
        elif existing == "erase":
            fo.delete_dataset(name)
            result = fo.Dataset(name)
        else:
            raise FileExistsError("Dataset already exists")
    else:
        result = fo.Dataset(name)
    # Manual casting to a dataset because the fiftyone Dataset constructor is
    # not compliant with standard metaclasses
    # see https://github.com/microsoft/pyright/discussions/5583
    result = cast(fo.Dataset, result)
    if bounding_box_formats is None:
        bounding_box_formats = {}
    if not isinstance(annotations_attributes_columns, dict):
        annotations_attributes_columns = {
            name: annotations_attributes_columns for name in annotations
        }

    images, image_tag_columns = make_fiftyone_compatible(images, image_tag_columns)
    for name in list(annotations.keys()):
        annotations[name], annotations_attributes_columns[name] = (
            make_fiftyone_compatible(
                annotations[name], annotations_attributes_columns[name]
            )
        )

    bbox_column_names = bbox_converter.column_names_from_format_string("xywh")
    fo_detections = {}
    for annotations_name, annotations_frame in annotations.items():
        if len(annotations_frame) == 0:
            continue
        input_format = bounding_box_formats.get(annotations_name, "XYWH")
        bbox = bbox_converter.convert_bbox(
            annotations_frame, images, input_format=input_format, output_format="xywh"
        )

        bbox = pd.Series(
            bbox[bbox_column_names].to_numpy().tolist(),
            index=annotations_frame.index,
            name="bbox",
        ).to_frame()
        to_concat = [annotations_frame, bbox]
        if label_map is not None and "category_str" not in annotations_frame.columns:
            to_concat.append(
                annotations_frame["category_id"]
                .replace(label_map)
                .rename("category_str")
                .to_frame()
            )

        current_converted_annotations = pd.concat(to_concat, axis=1)
        current_fo_detections = annotations_to_fiftyone(
            current_converted_annotations,
            annotations_attributes_columns.get(annotations_name, []),
            bbox_column="bbox",
            allow_keypoints=allow_keypoints,
        )
        current_fo_detections["image_id"] = annotations_frame["image_id"]
        fo_detections[annotations_name] = current_fo_detections

    samples = {}
    sample_ids = pd.Series(
        pd.NA,  # pyright: ignore
        index=images.index,
        name="sample_id",
    )
    detections_fo_metadata = {
        annotation_name: fo_detections_frame[["fo_id", "is_keypoint"]]
        for annotation_name, fo_detections_frame in fo_detections.items()
    }

    for image_id, image_data in tqdm(images.iterrows(), total=len(images)):
        image_path = images_root / image_data["relative_path"]
        metadata = fo.ImageMetadata(
            width=image_data["width"], height=image_data["height"]
        )
        sample = fo.Sample(
            filepath=image_path,
            lours_id=image_id,
            relative_path=str(image_data["relative_path"]),
            metadata=metadata,
            split=image_data.get("split", None),
            **image_data.loc[image_tag_columns].dropna(),  # pyright: ignore
        )
        samples[image_id] = sample
        sample_ids[image_id] = sample.id

    for annotation_name, fo_detections_frame in fo_detections.items():
        for (image_id, is_keypoint), detections_subset in fo_detections_frame.groupby(
            [
                "image_id",
                "is_keypoint",
            ]
        )["fo_detection"]:
            if is_keypoint:
                image_detections = fo.Keypoints(keypoints=detections_subset.to_list())
            else:
                image_detections = fo.Detections(detections=detections_subset.to_list())
            attribute_name = (
                f"{annotation_name}_{'keypoint' if is_keypoint else 'detection'}"
            )
            samples[image_id][attribute_name] = image_detections

    if do_update:
        result.merge_samples(samples.values(), key_field="relative_path", dynamic=True)
    else:
        result.add_samples(samples.values(), dynamic=True)
    result.add_dynamic_sample_fields()
    result.save()

    return (result, sample_ids, detections_fo_metadata)



import warnings
from collections.abc import Iterable

import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix as confusion_matrix_sklearn

from ...utils import BBOX_COLUMN_NAMES


def construct_matches_df(
    predictions_targets_df: pd.DataFrame, min_iou: float = 0
) -> pd.DataFrame:
    """From a dataframe with targets and predictions, all concatenated together,
    construct a list of match pairs between prediction and targets. Unmatched
    predictions or targets get a <NA> match id. Note that all bounding boxes in the
    input dataframe are assumed to be of the same category and the same image, the
    grouping must have already been done by the user before.

    Args:
        predictions_targets_df: DataFrame comprising target and prediction info must
            have the following columns:

             - ``groundtruth`` : bool value to know if it's a target or a prediction
             - ``box_x_min``, ``box_y_min``, ``box_width``, ``box_height``: Bounding box
               information to compute IoU

        min_iou: IoU above which the detection is considered valid.
            Note that the lower bound is not inclusive. Defaults to 0.

    Returns:
        DataFrame of matches. Will contain prediction_id and groundtruth_id columns.
        Index is irrelevant. Each prediction id and target id should appear once and
        only once. As such, at worse (no match at all), the dataframe will be N+M rows
        with N the number of predictions and M the number of targets

    """
    groundtruth = predictions_targets_df[predictions_targets_df["groundtruth"]]
    predictions = predictions_targets_df[~predictions_targets_df["groundtruth"]]
    ious = get_ious(groundtruth, predictions)
    detection_matches, groundtruth_matches = get_matches(
        ious, predictions["confidence"], min_iou
    )
    matches = detection_matches.reset_index(names="prediction_id").rename(
        columns={"match_id": "groundtruth_id"}
    )
    not_detected = groundtruth_matches[groundtruth_matches["match_id"].isna()]
    not_detected = not_detected.reset_index(names="groundtruth_id").rename(
        columns={"match_id": "prediction_id"}
    )
    return pd.concat([matches, not_detected])


def get_ious(groundtruth: pd.DataFrame, predictions: pd.DataFrame) -> pd.DataFrame:
    """From two dataframes of annotations, generate a matrix of iou of size N x M where
    N is the number of predictions and M is the number targets.
    Rows are sorted by prediction confidence

    Note that this does not check the category_id, only the bounding box coordinates.

    Next, encapsulate it in a dataframe with index and columns named after prediction
    and target ids.

    Args:
        groundtruth: DataFrame comprising bounding box targets data.
            Must include at least ``box_x_min``, ``box_y_min``, ``box_width``,
            ``box_height``
        predictions: DataFrame comprising bounding box prediction data.
            Must include same columns as groundtruth, plus the ``confidence`` column.

    Returns:
        DataFrame comprising iou values between groundtruth and predictions. Index is
        prediction id, column name is target id
    """
    # Extract bbox coordinates from groundtruth and pred. Note that prediction bbox data
    # is one more dimension to use array broadcasting
    # each array is of shape [M] (implicitly, [1, M])
    x1, y1, w1, h1 = groundtruth[BBOX_COLUMN_NAMES].values.T
    # each array is of shape [N, 1]
    x2, y2, w2, h2 = predictions[BBOX_COLUMN_NAMES].values.T[..., None]
    # Compute area of intersection
    # Here we use array broadcasting so that every constructed array is of size NxM
    xmin = np.maximum(x1, x2)
    xmax = np.minimum(x1 + w1, x2 + w2)
    ymin = np.maximum(y1, y2)
    ymax = np.minimum(y1 + h1, y2 + h2)
    area = (xmax - xmin) * (ymax - ymin)
    area[(xmax < xmin) | (ymax < ymin)] = 0
    ious = pd.DataFrame(
        area / (w1 * h1 + w2 * h2 - area),
        index=predictions.index,
        columns=groundtruth.index,
    )

    return ious


def get_matches(
    iou_df: pd.DataFrame, confidence: pd.Series | None = None, min_iou: float = 0
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get the best matching target for every prediction and
    return matching target (if any) for every prediction and
    matching prediction (if any) for every target
    Prediction are either reordered by confidence, or assumed already ordered in the
    first place.


    Args:
        iou_df: IoU values matrix encapsulated in a dataframe to index
            rows with prediction ids and columns with target ids
        confidence: series with the number of rows as iou_df, will
            be used to reorder iou_df's rows in descending order. If not given, will
            assume iou_df is already ordered.
        min_iou: Minimum IoU value above which a match is considered
            valid.

    Returns:
        dataframes of matching ids with corresponding
        ious. First df is indexed by prediction ids, second df is indexed by target id
    """
    if confidence is not None:
        ious = iou_df.reindex(confidence.sort_values(ascending=False).index)
    else:
        ious = iou_df.copy()
    # Note that we use the Int64 type, which is the regular int64 + NA value, which is
    # used here to designate the absence of match
    # Both matches dataframes are initialized to have zero match and will be iteratively
    # updated
    detection_matches = pd.DataFrame(
        np.zeros((len(ious), 2)),
        index=ious.index,
        columns=["iou", "match_id"],
    )
    groundtruth_matches = pd.DataFrame(
        np.zeros((len(ious.columns), 2)),
        index=ious.columns,
        columns=["iou", "match_id"],
    )
    detection_matches["match_id"] = pd.NA
    groundtruth_matches["match_id"] = pd.NA

    match_dtypes = {"iou": float, "match_id": "Int64"}
    detection_matches = detection_matches.astype(match_dtypes)
    groundtruth_matches = groundtruth_matches.astype(match_dtypes)

    # Iterative vectorize matching algorithm
    # 1 - Get best target match of each prediction
    # 2 - Remove every prediction and corresponding target until the first duplicate
    # 3 - Update aforementioned match dataframes accordingly
    # 4 - Repeat with this new subset
    # Note that we don't need to compute best target match each time (only until the
    # first duplicate), but that fact that it is vectorized across the iou matrix
    # makes it basically free.
    while len(ious) > 0:
        best_iou = ious.max(axis=1)
        valid = best_iou > min_iou
        ious = ious[valid]
        best_iou = best_iou[valid]
        if len(ious) == 0:
            break
        best_matches = ious.idxmax(axis=1)
        duplicated = best_matches.duplicated()
        if not duplicated.max():
            # No duplicate (max is False), perfect matching
            first_duplicated = len(duplicated)
        else:
            # Get first occurrence of duplicated == True
            first_duplicated = duplicated.argmax()

        # Partition between matched and not matched yet
        matched = best_matches.iloc[:first_duplicated]
        matched_iou = best_iou.iloc[:first_duplicated]
        not_matched = best_matches.iloc[first_duplicated:]

        ious = ious.loc[not_matched.index].drop(pd.Index(matched.values), axis=1)

        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            detection_matches.loc[matched.index, "match_id"] = matched
            groundtruth_matches.loc[matched, "match_id"] = matched.index.to_numpy()

        # Get corresponding iou values
        detection_matches.loc[matched.index, "iou"] = matched_iou
        groundtruth_matches.loc[matched, "iou"] = matched_iou

    return detection_matches, groundtruth_matches


def pr_curve(
    results: pd.DataFrame,
    min_iou: float = 0,
    betas: Iterable[float] = (1,),
    reindex_series: pd.Series | None = None,
) -> pd.DataFrame:
    r"""Construct Precision Recall curve from results dataframe and minimum iou below
    which detection is considered invalid

    Additionally, computes F-score with different :math:`\beta` values with the
    following equation.

    .. math:
        F_\beta = \frac{(1 + \beta^2)(\text{precision} \times \text{recall})}
                       {\text{precision} \times \beta^2 + \text{recall}}

    Args:
        results: Dataframe modelling detections, with corresponding
            confidence and groundtruth (whether this detection would be True positive or
            a False positive). Should include the columns ``groundtruth``, ``iou`` and
            ``confidence``, and rows should be sorted so that confidence values are
            sorted.
        min_iou: Value below which the detection is considered
            invalid. In other words, the groundtruth becomes ``False``. The prediction
            becomes a False Positive, and the corresponding groundtruth is a False
            negative. Defaults to 0.
        betas: beta values to compute the F-Score with. Must be an iterable of floats.
            Defaults to ``(1,)``
        reindex_series: Recall bins to reindex the curve. before returning it.

    Returns:
        Precision Recall curve dataframe. Columns are ``precision``,
        ``recall``, ``f{beta}_score`` and ``confidence_threshold``, where betas are the
        given :math:`\beta` values in ``betas`` (see equation above).
        Index is irrelevant.
    """
    results = results.sort_values("confidence", ascending=False)
    ntargets = results["groundtruth"].sum()
    confidence = results["confidence"].to_numpy()
    distinct_value_indices = np.diff(confidence).astype(bool)
    distinct_value_indices = np.append(distinct_value_indices, True)
    confidence = confidence[distinct_value_indices]
    # Cumulative sum of true positives, from which we only extract the maximum for
    # distinct confidence value
    tp_count = (results["groundtruth"] * (results["iou"] > min_iou)).to_numpy().cumsum()
    tp_count = tp_count[distinct_value_indices]

    # Precision and recall
    # Precision is true positive / number of positive predictions
    # Recall is true positive / number of total targets (even the ones with IOU of zero)
    precision = tp_count / (1 + distinct_value_indices.nonzero()[0])
    # In the degenerate case of no targets to be detected, the recall cannot be computed
    # Hence the NaN
    recall = tp_count / ntargets if ntargets > 0 else tp_count * np.NaN

    # Add 2 points for each extreme
    # Precision will not be above first value,
    # no matter how  high the confidence threshold is
    # Recall will not be above last value,
    # no matter how low the confidence threshold is
    # We still add the extremal points with recall = 1 and precision = 0 and
    # precision = 1 and recall = 0 for completeness
    # Wen this curve is reindexed by precision or recall
    # (which is the case for pycocotools).

    # Note pyright ignore flags to be removed as soon as we get the pre-commit hook
    # pyright 1.1.206
    # See https://github.com/microsoft/pyright/issues/2809
    precision = np.concatenate([[1], precision[:1], precision, [0, 0]])
    recall = np.concatenate([[0, 0], recall, recall[-1:], [1]])
    confidence = np.concatenate([[1, 1], confidence, [0, 0]])

    # Make sure the precision is only decreasing.
    # The rationale is that the true precision recall curve (thus with infinite number
    # of points) is only decreasing.
    # But the way it is constructed with a finite dataset makes precision drop when
    # A false positive occurs, and increase again at the next true positive.
    # Most conservative way of constructing a realistic curve is to make points of
    # dropping precision equal to the next highest precision.
    # For that we use numpy's universal function, and more specifically the accumulate
    # feature
    # see https://numpy.org/doc/stable/reference/generated/numpy.ufunc.accumulate.html
    precision = np.maximum.accumulate(precision[::-1])[::-1]
    f_scores = {}
    for beta in betas:
        # See https://en.wikipedia.org/wiki/F-score for formula
        f_scores[f"f{beta}_score"] = (
            (1 + beta**2) * (precision * recall) / (precision * beta**2 + recall + 1e-5)
        )
    result = pd.DataFrame(
        np.stack([precision, recall, confidence, *f_scores.values()], axis=1),
        columns=["precision", "recall", "confidence_threshold", *f_scores.keys()],
    ).fillna(0)

    # Remove points which are not useful to draw the curve or compute the mAP, ie the
    # points  that are on a straight line
    constant_precision = (result["precision"].diff() == 0) & (
        result["precision"].diff(-1) == 0
    )
    result = result[~constant_precision]

    constant_recall = (result["recall"].diff() == 0) & (result["recall"].diff(-1) == 0)
    result = result[~constant_recall]

    if reindex_series is not None:
        result = result.set_index(reindex_series.name)
        # Remove duplicated index values, otherwise reindex will error
        result = result[~result.index.duplicated(keep="last")]
        result = result.fillna(0)
        result = result.reindex(reindex_series, method="ffill").reset_index()
    return result


def compute_average_precision(pr_curve: pd.DataFrame) -> float:
    """Compute average precision from dataframe with precision and recall values.
    Precision values are averaged over recall values.

    Note:
        We compute the right Riemann sum, i.e. we only consider the value on the right
        for a particular recall interval.

    Args:
        pr_curve: Dataframe with ``precision`` and ``recall`` columns.

    Returns:
        Average precision for this particular PR curve
    """
    sorted_pr_curve = pr_curve.sort_values("recall")
    precision = sorted_pr_curve["precision"]
    # First value of recall_diff is NaN, replace it with 0 so that we discard
    # the first precision value
    recall_diff = sorted_pr_curve["recall"].diff().fillna(0)
    return (precision * recall_diff).sum()


def resample_count(
    original_confidences: Iterable[float], new_confidences: Iterable[float]
) -> pd.Series:
    """Take a sequence of confidence values and resample it assuming at each new
    original confdience value, one object is added.

    Result is the number of objects that would have been detected for each value in
    new confidence.

    Note:
        ``new_confidences`` must be sorted unique values.

    Args:
        original_confidences: Original set of confidence value. Each confidence value
            corresponds to one detected object.
        new_confidences: New set of confidence values to resample the number of detected
            objects from. Usually, a range of N elements, from 0 to 1.

    Returns:
        Series named ``count`` with the same length as ``new_confidences``, index set as
        ``new_confidences``, named ``confidence``, and values set to count values
        corresponding to confidence threshold given in the index.
    """
    counts = (
        pd.Series(list(original_confidences))
        .value_counts()
        .sort_index(ascending=False)
        .cumsum()
    )
    new_confidences = pd.Index(new_confidences, name="confidence")
    resampled = counts.reindex(new_confidences, method="ffill").fillna(0)

    return resampled


def confusion_matrix(matches: pd.DataFrame) -> pd.DataFrame:
    """Compute the confusion matrix for a given DataFrame.

    Args:
        matches: DataFrame containing the matches between groundtruth and predictions
            in which we expect to have the following columns :

            - ``prediction_label``
            - ``groundtruth_label``

            corresponding to the predicted and groundtruth labels,
            respectively, in order to compute the confusion matrix.

    Returns:
        A confusion matrix as DataFrame with class names as column names and row ids.
    """
    y_pred = matches["prediction_label"].fillna("None").astype(str)
    y_true = matches["groundtruth_label"].fillna("None").astype(str)

    # Create a list of all possible classes
    all_classes = sorted(set(y_pred) | set(y_true))
    if "None" in all_classes:
        all_classes.remove("None")
        all_classes.append("None")

    cm = confusion_matrix_sklearn(y_true, y_pred, labels=all_classes, normalize="true")

    return pd.DataFrame(
        cm, index=pd.Index(all_classes, name="label"), columns=all_classes
    )


def display_confusion_matrix(confusion_matrix: pd.DataFrame, title: str = ""):
    """Display a ConfusionMatrixDisplay object for a given Dataframe.

    Args:
        confusion_matrix: Dataframe containing the confusion matrix data
            as computed by :func:`.confusion_matrix`
        title: Confusion matrix's title

    """
    try:
        import matplotlib.pyplot as plt
        from sklearn.metrics import ConfusionMatrixDisplay
    except ImportError as e:
        raise ImportError(
            "Plotting libraries could not be loaded, make sure you have installed"
            " Lours with the 'plot-utils' extra"
        ) from e

    display_labels = confusion_matrix.columns.to_list()

    # scaling text inside the matrix cells.
    # Text is not scaled according to the number of labels.
    # We need to make the text smaller if the matrix cells are getting smaller as well
    # somehow, the size of the cell is both inversly proportional to the number of
    # labels, and also gets smaller if the longest label is very long.
    # this algorithm tries to find the right font size, from xx-small to regular
    text_kw = {}
    n_labels = len(display_labels)
    max_label_length = max(map(len, display_labels))
    n_labels += max_label_length / 3
    if n_labels > 15:
        text_kw["fontsize"] = "xx-small"
    elif n_labels > 11:
        text_kw["fontsize"] = "x-small"
    elif n_labels > 9:
        text_kw["fontsize"] = "small"

    plot = ConfusionMatrixDisplay(
        confusion_matrix.values.round(2),
        display_labels=display_labels,
    ).plot(text_kw=text_kw)
    # Use maptlotlib's tick labels function so that we can rotation around the tick
    # and not around the label center. Otherwise, long labels might end up overlapping
    # the next labels
    plot.ax_.set_xticklabels(
        display_labels,
        rotation=45,
        ha="right",
        rotation_mode="anchor",
        fontsize="small" if max_label_length > 10 else "medium",
    )
    if max_label_length > 10:
        plot.ax_.set_yticklabels(
            display_labels,
            fontsize="small",
        )
    plt.title(title)
    plt.xlabel("Prediction label")
    plt.ylabel("Grounthruth label")
    plt.tight_layout()



"""Module dedicated to Dataset indexers, to be able to index Dataset with pandas style
loc and iloc methods
"""

from typing import Any, Generic, Literal, TypeVar

import pandas as pd

from lours.dataset import Dataset

D = TypeVar("D", bound=Dataset)


class DatasetImLocator(Generic[D]):
    """Locator class dedicated to index a dataset by its images as if we used pandas
    indexing methods on ``dataset.images`` and filtered annotations accordingly.

    Usually used in the context of :meth:`.Dataset.loc` and :meth:`.Dataset.iloc`
    """

    def __init__(self, dataset: D, mode: Literal["loc", "iloc"] = "loc") -> None:
        """Constructor

        Args:
            dataset: Dataset object to index
            mode: whether to use ``dataset.images.loc`` or ``dataset.images.iloc``.
                Defaults to "loc"
        """
        self.dataset = dataset
        self.mode = mode

    def __getitem__(self, index: Any) -> D:
        """Index the dataset with an index applied on ``dataset.images``

        Args:
            index: index object expected to be compatible with ``dataset.images.loc`` or
                ``dataset.images.iloc``.

        Returns:
            Indexed sub-dataset
        """
        if self.mode == "loc":
            new_images = self.dataset.images.loc[index]
        else:
            new_images = self.dataset.images.iloc[index]
        if isinstance(new_images, pd.Series):
            # The Series object indicates only one row. Convert it back to a frame
            new_images = new_images.to_frame().T.astype(self.dataset.images.dtypes)
        new_annotations = self.dataset.annotations.loc[
            self.dataset.annotations["image_id"].isin(new_images.index)
        ]
        return self.dataset.from_template(
            images=new_images,
            annotations=new_annotations,
            reset_booleanized=False,
        )


class DatasetAnnotLocator(Generic[D]):
    """Locator class dedicated to index a dataset by its annotations as if we used
    pandas indexing methods on ``dataset.annotations`` and potentially filtered
    empty images accordingly.

    Usually used in the context of :meth:`.Dataset.loc_annot` and
    :meth:`.Dataset.iloc_annot`
    """

    def __init__(
        self,
        dataset: D,
        mode: Literal["loc", "iloc"] = "loc",
        remove_emptied_images: bool = False,
    ) -> None:
        """Constructor

        Args:
           dataset: Dataset object to index
           mode: whether to use ``dataset.annotations.loc`` or
               ``dataset.annotations.iloc``. Defaults to "loc"
           remove_emptied_images: If se to True, will remove images that no longer have
               annotations, but will keep images that were already empty before.
               Defaults to False
        """
        self.dataset = dataset
        self.mode = mode
        self.remove_emptied_images = remove_emptied_images

    def __getitem__(self, index: Any) -> D:
        """Index the dataset with an index applied on ``dataset.annotations``

        Args:
            index: index object expected to be compatible with
                ``dataset.annotations.loc`` or ``dataset.annotations.iloc``.

        Returns:
            Indexed sub-dataset
        """
        if self.mode == "loc":
            new_annotations = self.dataset.annotations.loc[index]
        else:
            new_annotations = self.dataset.annotations.iloc[index]
        if isinstance(new_annotations, pd.Series):
            new_annotations = new_annotations.to_frame().T.astype(
                self.dataset.annotations.dtypes
            )
        if self.remove_emptied_images:
            already_empty_images = ~self.dataset.images.index.isin(
                self.dataset.annotations["image_id"]
            )
            already_empty_images_ids = self.dataset.images.index[
                already_empty_images
            ].tolist()
            new_images = self.dataset.images.loc[
                already_empty_images_ids + new_annotations["image_id"].unique().tolist()
            ]
        else:
            new_images = self.dataset.images
        return self.dataset.from_template(
            images=new_images,
            annotations=new_annotations,
            reset_booleanized=False,
        )



from collections.abc import Iterable
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from ...utils.grouper import (
    get_group_names,
    group_list,
    group_relational_data,
    groups_to_list,
)
from .detection_evaluator_base import DetectionEvaluatorBase
from .util import resample_count

if TYPE_CHECKING:
    pass


class CrowdDetectionEvaluator(DetectionEvaluatorBase):
    """Class specialization for crowd detection and counting tasks.
    Note that the constructor is the same as the base Evaluator

    See Also:
        :ref:`related tutorial </notebooks/4_demo_evaluation_crowd.ipynb>`
    """

    def compute_count_error(
        self,
        groups: group_list = "category_id",
        quantiles: Iterable[float] = (0.25, 0.5, 0.75),
        confidence_index: Iterable[float] | None = None,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Compute Count error metrics, both absolute (in number of objects found) and
        relative (with respect to groundtruth number of objects) with respect to
        confidence threshold.

        Along with these metrics, it computes standard deviation of absolute/relative
        error and quantiles for error values.

        See Also:
            :ref:`related tutorial </notebooks/4_demo_evaluation_crowd.ipynb>`

        Computed metrics:
            - Mean Absolute Error (MAE)
            - Root of Mean Square Error (RMSE)
            - Mean Relative Error (MRE)
            - Root of Mean Square Relative Error (RMSRE)

        Args:
            groups: Groups of image or annotation attributes to use to
                partition evaluation results to compute multiple PR curves. Must be a
                :obj:`.group_list` . Defaults to "category_id".
            quantiles: quantile values to get with respect to confidence threshold,
                aggregated per image. Must contain the median value (i.e. 0.5).
                Defaults to (0.25, 0.5, 0.75).
            confidence_index: sequence of confidence thresholds to compute the metric
                on. If set to None, will be 101 equidistant points, from 0 to 1.
                Defaults to None.

        Returns:
            A pair of DataFrames.

            - a DataFrame with computed metrics, with multiindex columns, for
              absolute and relative metrics with respect to confidence
            - a DataFrame with detailed error values with respect to confidence for each
              image, in order to compute statistics manually.

        Example:
            >>> from lours.utils.doc_utils import dummy_dataset
            >>> groundtruth = dummy_dataset(
            ...     10, 1000, label_map={0: "person", 1: "car"}, keypoints_share=1
            ... )
            >>> predictions = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     keypoints_share=1,
            ...     add_confidence=True,
            ... )
            >>> evaluator = CrowdDetectionEvaluator(
            ...     groundtruth=groundtruth, predictions=predictions
            ... )
            >>> errors, detailed = evaluator.compute_count_error()
            >>> errors
                                   absolute              ...  relative
                                        MAE        RMSE  ...     q0.75        model
            category_id confidence                       ...
            0           0.00          452.0  452.378824  ...  9.913239  predictions
                        0.01          447.8  448.170057  ...  9.779314  predictions
                        0.02          442.3  442.670645  ...  9.655792  predictions
                        0.03          437.6  437.945887  ...  9.569740  predictions
                        0.04          433.5  433.856774  ...  9.412411  predictions
            ...                         ...         ...  ...       ...          ...
            1           0.96           34.2   35.883144  ... -0.540094  predictions
                        0.97           38.0   39.549968  ... -0.630391  predictions
                        0.98           42.6   43.395852  ... -0.768797  predictions
                        0.99           46.3   46.764303  ... -0.911782  predictions
                        1.00           50.7   51.073476  ... -1.000000  predictions
            <BLANKLINE>
            [202 rows x 14 columns]

            Get the confidence threshold where the Mean Average Error is the lowest,
            and show the corresponding rows (one per category).

            >>> mae = errors[("absolute", "MAE")]
            >>> mae
            category_id  confidence
            0            0.00          452.0
                         0.01          447.8
                         0.02          442.3
                         0.03          437.6
                         0.04          433.5
                                       ...
            1            0.96           34.2
                         0.97           38.0
                         0.98           42.6
                         0.99           46.3
                         1.00           50.7
            Name: (absolute, MAE), Length: 202, dtype: float64
            >>> best_mae = errors.loc[mae.groupby(level=0).idxmin()]
            >>> best_mae
                                   absolute            ...  relative
                                        MAE      RMSE  ...     q0.75        model
            category_id confidence                     ...
            0           0.89            5.4   7.655064  ...  0.116153  predictions
            1           0.88           10.9  13.939153  ...  0.310000  predictions
            <BLANKLINE>
            [2 rows x 14 columns]
            >>> best_mae.reset_index().iloc[0]
            category_id                     0
            confidence                   0.89
            absolute     MAE              5.4
                         RMSE        7.655064
                         std         7.788881
                         q0.25            0.0
                         q0.50            2.5
                         q0.75            5.5
                         model    predictions
            relative     MRE          0.10748
                         RMSRE       0.145579
                         std         0.145805
                         q0.25            0.0
                         q0.50       0.055717
                         q0.75       0.116153
                         model    predictions
            Name: 0, dtype: object
        """

        def add_image_id(g: list) -> list:
            if "image_id" not in g:
                return [*g, "image_id"]
            return g

        groups = groups_to_list(groups)
        group_names = get_group_names(groups)
        gt_group_dict, *_ = group_relational_data(self.groundtruth, groups, self.images)
        gt_pandas_groups = [gt_group_dict[name] for name in group_names]
        gt_count = (
            self.groundtruth.groupby(add_image_id(gt_pandas_groups))
            .size()
            .rename("gt_count")  # pyright: ignore
        )
        mae_curves = []
        detailed_error_counts = []
        for (
            current_predictions_name,
            current_predictions_frame,
        ) in self.predictions_dictionary.items():
            if confidence_index is None:
                current_confidence_index = np.linspace(0, 1, 101)
            else:
                current_confidence_index = confidence_index

            group_dict, *_ = group_relational_data(
                current_predictions_frame, groups, self.images
            )
            pandas_groups = [group_dict[name] for name in group_names]
            tqdm.pandas()
            prediction_counts = (
                current_predictions_frame.groupby(add_image_id(pandas_groups))[
                    "confidence"
                ]
                .progress_apply(  # pyright: ignore
                    partial(resample_count, new_confidences=current_confidence_index)
                )
                .rename("count")
                .to_frame()
            )
            prediction_counts = prediction_counts.join(gt_count).fillna(0)
            prediction_counts["error"] = (
                prediction_counts["count"] - prediction_counts["gt_count"]
            )
            prediction_counts["rel_error"] = (
                prediction_counts["error"] / prediction_counts["gt_count"]
            )
            prediction_counts["abs_error"] = prediction_counts["error"].abs()
            prediction_counts["abs_rel_error"] = prediction_counts["rel_error"].abs()
            prediction_counts["sq_error"] = prediction_counts["error"].pow(2)
            prediction_counts["sq_rel_error"] = prediction_counts["rel_error"].pow(2)
            prediction_counts["model"] = current_predictions_name
            detailed_error_counts.append(prediction_counts)

            grouped = prediction_counts.groupby([*group_names, "confidence"])
            mae = grouped["abs_error"].mean().rename("MAE")
            rmse = np.sqrt(grouped["sq_error"].mean()).rename("RMSE")
            mre = grouped["abs_rel_error"].mean().rename("MRE")
            rmsre = np.sqrt(grouped["sq_rel_error"].mean().rename("RMSRE"))

            def q_at(y):
                def q(x):
                    return x.quantile(y)

                q.__name__ = f"q{y:0.2f}"
                return q

            stat_agg_functions = ["std", *[q_at(q) for q in quantiles]]
            stats = grouped["error"].agg(stat_agg_functions)
            rel_stats = grouped["rel_error"].agg(stat_agg_functions)
            absolute_result = pd.concat([mae, rmse, stats], axis=1)
            relative_result = pd.concat([mre, rmsre, rel_stats], axis=1)
            absolute_result["model"] = relative_result["model"] = (
                current_predictions_name
            )
            result = pd.concat(
                [absolute_result, relative_result],
                axis=1,
                keys=["absolute", "relative"],
            )
            mae_curves.append(result)
        mae_curves = pd.concat(mae_curves)
        detailed_error_counts = pd.concat(detailed_error_counts)
        return mae_curves, detailed_error_counts

    def compute_normalized_precision_recall(self) -> pd.DataFrame:
        """Compute nAP between detected points and ground truth according to the
        algorithm proposed in [Ref]_

        .. [Ref] Song, Q., Wang, C., Jiang, Z., Wang, Y., Tai, Y., Wang, C. & Wu, Y.
            Rethinking counting and localization in crowds: A purely point-based
            framework.
            2021 IEEE/CVF International Conference on Computer Vision (pp. 3365-3374).
            https://openaccess.thecvf.com/content/ICCV2021/html/Song_Rethinking_Counting_and_Localization_in_Crowds_A_Purely_Point-Based_Framework_ICCV_2021_paper.html
        """
        raise NotImplementedError



import json
import shutil
from collections.abc import Iterable
from datetime import date
from pathlib import Path

import pandas as pd
from imageio.v3 import imread, imwrite

from lours.utils import BBOX_COLUMN_NAMES
from lours.utils.bbox_converter import (
    column_names_from_format_string,
    export_bbox,
    import_bbox,
)
from lours.utils.testing import assert_images_valid

from ..dataset import Dataset
from .common import parse_annotation_name


def from_coco(
    coco_json: Path | str,
    images_root: Path | str | None = None,
    dataset_name: str | None = None,
    split: str | None = None,
    label_map: dict[int, str] | None = None,
    box_format: str = "XYWH",
    drop_columns: Iterable[str] = ("iscrowd", "segmentation"),
) -> Dataset:
    """Load a coco json file into a dictionary.
    Note that there is only one split per file, which needs to be given by caller.
    See `specifications`__ (only Object detection)

        .. __: https://cocodataset.org/#format-data

    Notes:
        - ``from_coco`` is compatible with bounding box annotations without
          ``category_id`` field, but then you will need to have a label map of only one
          entry, which will be assigned to every bounding box.
        - If split value is not given, it will try to deduce it from the file name.
          More specifically, it will search a ``<name>_<split>.json`` pattern and assign
          ``name`` to the dataset name and ``split`` to the split value.

    Args:
        coco_json: path of json file
        images_root: folder which file_name of images are relative to
        dataset_name: If specified, will be the dataset name, used when showing the
            dataset or exporting in other formats such as fiftyone. If not specified,
            the dataset name will be deduced from the name of the json file.
        split: split of given json file.
            If not set, will try to deduce from filename. Defaults to None.
        label_map: Optional dictionary to specify the name of each category id. If not
            set, will try to deduce it from the json itself, in the field `categories`
            at its root.
        box_format: what type of annotation the json file will have.
            It will be converted back to XYWH. Defaults to XYWH
        drop_columns: list of names of columns that need to be dropped from the parsed
            json dictionary.

    Returns:
        Loaded dataset object
    """
    # If given paths are string, convert them to Path
    coco_json = Path(coco_json)
    if images_root is not None:
        images_root = Path(images_root)
    else:
        images_root = coco_json.parent
    parsed_dataset_name, parsed_split = parse_annotation_name(
        annotations_file_path=coco_json
    )
    if dataset_name is None:
        dataset_name = parsed_dataset_name
    if split is None:
        split = parsed_split
    with open(coco_json) as f:
        coco_annotations = json.load(f)

    images = pd.json_normalize(coco_annotations["images"]).set_index("id")
    images["relative_path"] = images["file_name"].apply(Path)  # pyright: ignore
    images["type"] = images["relative_path"].apply(lambda x: x.suffix)
    if split is not None:
        images["split"] = split
    images = images.drop("file_name", axis=1)

    if label_map is None:
        try:
            label_map = {c["id"]: c["name"] for c in coco_annotations["categories"]}
        except KeyError:
            label_map = {}

    try:
        annotations = pd.json_normalize(coco_annotations["annotations"]).set_index("id")
    except KeyError:
        annotations = pd.DataFrame(
            [],
            columns=[
                *column_names_from_format_string("XYWH"),
                "image_id",
                "category_id",
            ],
        )
        annotations.index.name = "id"
        return Dataset(images_root, images, annotations, label_map)
    # Don't deal with iscrowd=1 for now
    if "iscrowd" in annotations.columns:
        annotations = annotations[annotations["iscrowd"] == 0]
    bboxes = pd.DataFrame(
        list(annotations["bbox"]),
        index=annotations.index,
        columns=column_names_from_format_string(box_format),
    )
    bboxes = import_bbox(
        bboxes, images, image_ids=annotations["image_id"], input_format=box_format
    )
    annotations = pd.concat([annotations, bboxes], axis=1)
    annotations = annotations.drop([*drop_columns, "bbox"], axis=1, errors="ignore")

    if "category_id" not in annotations.columns:
        assert len(label_map) == 1
        annotations["category_id"] = list(label_map)[0]

    if annotations["category_id"].hasnans:
        raise ValueError(
            "Some category ids in annotations are undefined. Make sure either every"
            " annotation in your coco file has a `category_id` field, or none of them"
            " have"
        )

    if "score" in annotations:
        annotations = annotations.rename(columns={"score": "confidence"})
    return Dataset(images_root, images, annotations, label_map, dataset_name)


def from_coco_keypoints(
    coco_json: Path | str,
    images_root: Path | str | None = None,
    dataset_name: str | None = None,
    split: str | None = None,
    box_format: str = "XY",
    category_name: str | None = "head",
):
    """Special coco loading function for crowds, it will assume point box format
    (either XY or xy), only one category, with an id of 0, and a category name of person
    (unless specified otherwise in the coco format)

    Args:
        coco_json: path of json file
        images_root: folder which file_name of images are relative to
        dataset_name: If specified, will be the dataset name, used when showing the
            dataset or exporting in other formats such as fiftyone. If not specified,
            the dataset name will be deduced from the name of the json file.
        split: split of given json file.
            If not set, will try to deduce from filename. Defaults to None.
        box_format: what type of annotation the json file will have.
            It will be converted back to XYWH, with box width and height set to 0.
            Defaults to XY
        category_name: name of the only category of this coco json file.
            It will then call the ``from_coco`` original version with a label map option
            set to ``{0: category_name}``.
            If set to None, will deduce it from coco file. Defaults to "person".

    Returns:
        Loaded dataset object
    """
    return from_coco(
        coco_json=coco_json,
        images_root=images_root,
        dataset_name=dataset_name,
        split=split,
        box_format=box_format,
        label_map={0: category_name} if category_name is not None else None,
    )


def dataset_to_coco(
    dataset: Dataset,
    output_path: Path | str,
    copy_images: bool = False,
    to_jpg: bool = True,
    overwrite_images: bool = False,
    overwrite_labels: bool = False,
    add_split_suffix: bool | None = None,
    box_format: str = "XYWH",
    version: str = "0",
    contributor: str = "XXII",
) -> None:
    """Save dataset to COCO format. Note that by default, no image or image path is
    manipulated

    Args:
        dataset: Dataset object to save
        output_path: output folder where to save the json file,
            and optionally the images. Can also be the name of the output json file when
            there is only one split value.
        copy_images: If True, will copy images linked by annotations in
            a "data" folder, similar to 51. Defaults to False.
        to_jpg: if True, along with previous option, will convert
            images to jpg if needed. Defaults to True.
        overwrite_images: if False with copy_images True,
            will skip images that are already copied. Defaults to True.
        overwrite_labels: if False,
            will skip annotation that are already created. Defaults to True.
        add_split_suffix: if True, will add the split name to the
            name of the json file. cannot be False if dataset has multiple splits.
            If not set, will add suffix only if dataset has multiple splits.
        box_format: what type of annotation the json file will have.
            It will be converted from XYWH. Defaults to XYWH
        version: Arbitrary version number for dataset metadata.
            Defaults to "0".
        contributor: Arbitrary contributor info for dataset metadata.
            Defaults to "XXII".
    """
    if copy_images:
        try:
            assert_images_valid(dataset, load_images=False)
        except AssertionError as e:
            raise ValueError(
                "Dataset images are missing, check that the images root folder is the"
                " right one"
            ) from e
    output_path = Path(output_path).resolve()
    dataset = dataset.debooleanize()
    if output_path.suffix == ".json":
        output_file_name = output_path.stem
        output_path = output_path.parent
    else:
        output_path.mkdir(exist_ok=True, parents=True)
        output_file_name = "annotations"
    now = date.today()

    if add_split_suffix is None:
        add_split_suffix = len(dataset.annotations.split.unique()) > 1

    if not add_split_suffix:
        assert (
            "split" not in dataset.annotations
            or len(dataset.annotations.split.unique()) == 1
        ), "Cannot remove split suffix because dataset has more than one split"

    if copy_images:
        output_img_paths = dataset.images["relative_path"].apply(
            lambda x: output_path / "data" / x
        )
        if to_jpg:
            output_img_paths = output_img_paths.apply(lambda x: x.with_suffix(".jpg"))
        for i, row in dataset.images.iterrows():
            input_img_path = (dataset.images_root / row["relative_path"]).resolve()
            output_img_path = output_img_paths.loc[i]  # pyright: ignore
            if overwrite_images or not output_img_path.is_file():
                output_img_path.parent.mkdir(parents=True, exist_ok=True)
                if not to_jpg or row["type"].lower() in [".jpg", ".jpeg"]:
                    shutil.copy(input_img_path, output_img_path)
                else:
                    image = imread(input_img_path)
                    imwrite(output_img_path, image[..., :3])

    if copy_images and to_jpg:
        image_paths_str = dataset.images["relative_path"].apply(
            lambda x: str(x.with_suffix(".jpg"))
        )
    else:
        image_paths_str = dataset.images["relative_path"].apply(str)
    converted_images = dataset.images.drop(["relative_path", "type"], axis=1)
    converted_images["file_name"] = image_paths_str
    converted_bbox = export_bbox(
        dataset.annotations, dataset.images, output_format=box_format
    )
    converted_bboxes = pd.Series(
        converted_bbox[column_names_from_format_string(box_format)].to_numpy().tolist(),
        index=dataset.annotations.index,
        name="bbox",
    )
    converted_annotations = dataset.annotations.drop(
        [*BBOX_COLUMN_NAMES, "category_str"], axis=1
    )
    if "confidence" in converted_annotations:
        converted_annotations = converted_annotations.rename(
            columns={"confidence": "score"}
        )
    converted_annotations = pd.concat([converted_annotations, converted_bboxes], axis=1)
    converted_annotations["iscrowd"] = 0

    if "split" not in converted_images:
        converted_images["split"] = None
    if "split" not in converted_annotations:
        converted_annotations["split"] = dataset.images.loc[
            converted_annotations["image_id"], "split"
        ]

    for split_name, split_imgs in converted_images.groupby("split", dropna=False):
        if add_split_suffix and not pd.isna(split_name):
            output_json_path = output_path / f"{output_file_name}_{split_name}.json"
        else:
            output_json_path = output_path / f"{output_file_name}.json"
        if not overwrite_labels and output_json_path.is_file():
            continue
        print(f"saving {split_name} to {output_json_path.name}...")
        info_dict = {
            "description": "Made with XXII dataset helper",
            "url": "https://xxii.fr",
            "version": version,
            "year": now.year,
            "contributor": contributor,
        }
        images_list = split_imgs.reset_index().to_dict("records")
        if pd.isna(split_name):
            split_annotations = converted_annotations[
                converted_annotations["split"].isna()
            ]
        else:
            split_annotations = converted_annotations[
                converted_annotations["split"] == split_name
            ]
        print(split_annotations)
        print(split_annotations.keys())
        annotations_list = split_annotations.reset_index().to_dict("records")

        categories_list = [
            {
                "supercategory": "",
                "id": cat_id,
                "name": cat_name,
            }  # Not implemented yet,
            for cat_id, cat_name in dataset.label_map.items()
        ]
        coco_dict = {
            "info": info_dict,
            "images": images_list,
            "annotations": annotations_list,
            "categories": categories_list,
        }
        with open(output_json_path, "w") as f:
            json.dump(coco_dict, f, indent=2)



import os
from hashlib import sha256
from pathlib import Path

import pytest
from imageio import imread
from pytest_regressions.common import perform_regression_check

from lours.dataset import Dataset, from_parquet

from .testing import assert_dataset_equal


class DataSetRegressionFixture:
    """Lours Dataset Regression fixture implementation used on dataset_regression
    fixture.

    Loosely based on pytest-regression's DataFrameRegression.
    See `Related code on github`__

        .. __: https://github.com/ESSS/pytest-regressions/blob/master/src/pytest_regressions/dataframe_regression.py
    """

    def __init__(
        self, datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
    ) -> None:
        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self._force_regen = False
        self._with_test_class_names = False

    def check_fn(
        self, obtained_dataset_path: Path, expected_dataset_path: Path
    ) -> None:
        """Check that two saved parquet datasets are the same

        Args:
            obtained_dataset_path: path to current dumped dataset
            expected_dataset_path: path to saved dataset to compare the current
                object with.
        """
        actual_obtained_path = obtained_dataset_path.parent / (
            obtained_dataset_path.name + ".d"
        )
        actual_expected_path = expected_dataset_path.parent / (
            expected_dataset_path.name + ".d"
        )
        dataset1 = from_parquet(actual_obtained_path)
        dataset2 = from_parquet(actual_expected_path)

        assert_dataset_equal(dataset1, dataset2)

    def dump_fn(
        self, dataset: Dataset, filename: Path, with_image_checksum: bool = False
    ) -> None:
        """Function used to dump the dataset.

        Note:
            Since pytest-regressions only accepts files and :func:`.to_parquet()`
            method creates a folder, the dump function here will create an empty file
            and dump everything in a folder named after the file path with the ".d"
            added to it.

        Args:
            dataset: dataset to save
            filename: path to file that will be considered the file of the dataset.
                the actual dataset will be saved in the same path but with ".d" appended
                to it.
            with_image_checksum: If set to True, will load all images to get a checksum
                for each image based on its pixels, and save it in a dedicated column in
                the dataset's image dataframe.
        """
        output_folder = filename.parent / (filename.name + ".d")
        if with_image_checksum:
            checksums = {}
            for image_id, relative_path in dataset.images["relative_path"].items():
                image_path = dataset.images_root / relative_path
                image = imread(image_path)
                checksums[image_id] = sha256(image.tobytes()).hexdigest()
            dataset = dataset.from_template(
                images=dataset.images.assign(checksum=checksums)
            )
        dataset.to_parquet(output_folder, overwrite=True)
        # If dump was successful, make an empty file so that filename is a path to a
        # valid file, making pytest-regression happy
        filename.touch()

    def check(
        self,
        dataset: Dataset,
        basename: str | None = None,
        fullpath: "os.PathLike[str] | None" = None,
        check_images: bool = False,
    ) -> None:
        """Checks a dataset object, against a previously recorded version, or generate a
        new parquet version of it.

        Note:
            Since pytest-regressions only accepts files and :func:`.to_parquet()`
            method creates a folder, the dump function here will create an empty file
            and dump everything in a folder named after the file path with the ".d"
            added to it.


        Args:
            dataset: Dataset containing annotations for regression check
            basename: basename of the file to test/record. If not given the name
                of the test is used.
            fullpath: complete path to use as a reference file. This option
                will ignore ``embed_data`` completely, being useful if a reference
                file is located in the session data dir for example. Defaults to None.
            check_images: If set to True, will load all images to get a checksum of each
                image, based on its pixels, and save it in a dedicated column in
                the dataset's image dataframe. Defaults to False
        """
        import functools

        assert isinstance(dataset, Dataset), (
            "Only Lours datasets are supported on dataframe_regression fixture.\n"
            f"Object with type {type(dataset)} was given."
        )

        dump_fn = functools.partial(
            self.dump_fn, dataset, with_image_checksum=check_images
        )

        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=self.check_fn,
            dump_fn=dump_fn,
            extension="",
            basename=basename,
            fullpath=fullpath,
            force_regen=self._force_regen,
            with_test_class_names=self._with_test_class_names,
        )


@pytest.fixture
def dataset_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "DataSetRegressionFixture":
    """Regression fixture for datasets. Will save the dataset to parquet and compare
    with the already saved parquet dataset

    Args:
        datadir: Where to save the dataset in parquet format. If not set, will use the
            default value from pytest-data, i.e. folder with the name of the python
            file + file with the name of the test function.
        original_datadir: Where to load the original dataset, in parquet format. If not
            set, will use the default value from pytest-data, i.e. folder with the name
            of the python file + file with the name of the test function.
        request: pytest request object used by pytest-regression to get options from
            pytest shell command.

    Returns:
        Fixture used by pytest to dump a dataset, and compare its dump with the
        already existing one.

    Example:

        >>> from lours.utils.doc_utils import dummy_dataset
        >>> def testSomeData(dataset_regression):
        ...     my_dataset = dummy_dataset()
        ...     dataset_regression.check(my_dataset)
        ...
    """
    return DataSetRegressionFixture(datadir, original_datadir, request)



from collections.abc import Iterable
from pathlib import Path
from warnings import warn

from ..dataset import Dataset
from .common import get_image_info, get_images_from_folder, to_dataset_object


def from_folder(
    images_root: str | Path,
    split: str | None = "eval",
    label_map: dict[int, str] | None = None,
    dataset_path: str | Path | None = None,
) -> Dataset:
    """Load a folder of images into a dataset without annotations.

    Globbed image file formats are the following:

     - ".bmp"
     - ".dng"
     - ".jpeg"
     - ".jpg"
     - ".mpo"
     - ".png"
     - ".tif"
     - ".tiff"
     - ".webp"
     - ".pfm"

    Args:
        images_root: Root of folder to get images from
        split: Split value to apply to resulting dataset. Defaults to "eval".
        label_map: Optional label map to apply to dataset. Defaults to None.
        dataset_path: Deprecated name for images_root, if not None, triggers a warning
            and will be removed in future releases

    Returns:
        Dataset with images of given folder, but without annotations.
    """
    if dataset_path is not None:
        warn(
            "Dataset.dataset_path is deprecated in favor of Dataset.images_root",
            DeprecationWarning,
        )
        images_root = dataset_path
    images_root = Path(images_root)
    images_path = get_images_from_folder(images_root)
    if not images_path:
        warn(f"No image file found in folder {images_root}", RuntimeWarning)
    images = []
    for i, img_path in enumerate(images_path):
        image_data = get_image_info(i, img_path, images_root / img_path)
        image_data["split"] = split
        images.append(image_data)
    return to_dataset_object(
        images_root=images_root,
        label_map={} if label_map is None else label_map,
        images=images,
        annotations=[],
    )


def from_files(
    images_root: str | Path = "",
    file_names: str | Path | Iterable[str | Path] = "",
    split: str | None = "eval",
    label_map: dict[int, str] | None = None,
) -> Dataset:
    """Load a list of image paths into a dataset without annotations

    Note:
        Image paths can be globbing patterns as well. As such, if your folder only
        contains .jpg and .png files, calling this function with ``file_names`` set to
        ``["*.jpg", "*.jpeg"]`` will produce the same result as :func:`.from_folder`

    Note:
        Calling this function with ``file_names`` set to "*" is NOT equivalent to
        :func:`.from_folder` because the pattern used in :func:`.from_folder` is more
        constrained.

    Args:
        images_root: Root of folder to get images from. Defaults to "".
        file_names: files to add to the dataset. Can be paths or globbing pattern,
            must be relative to ``images_root``. Defaults to "".
        split: Split value to apply to resulting dataset. Defaults to "eval".
        label_map: Optional label map to apply to dataset. Defaults to None.

    Returns:
        Dataset with given images, but without annotations.
    """
    images_root = Path(images_root)
    if isinstance(file_names, str | Path):
        file_names = [file_names]
    total_image_paths = []
    for f in file_names:
        total_image_paths.extend(images_root.glob(str(f)))
    if not total_image_paths:
        warn(
            f"No image file found in folder {images_root} for the given file paths /"
            " patterns",
            RuntimeWarning,
        )
    images = []
    for i, img_path in enumerate(total_image_paths):
        image_data = get_image_info(i, img_path.relative_to(images_root), img_path)
        image_data["split"] = split
        images.append(image_data)
    return to_dataset_object(
        images_root=images_root,
        label_map={} if label_map is None else label_map,
        images=images,
        annotations=[],
    )



import socketserver
from argparse import ArgumentParser
from logging import warn
from pathlib import Path
from time import sleep
from typing import TYPE_CHECKING

from watchdog.events import (
    EVENT_TYPE_CLOSED,
    EVENT_TYPE_OPENED,
    FileSystemEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer
from watchdog.observers.api import EventQueue

from lours.dataset import Dataset, from_caipy, from_caipy_generic
from lours.utils import try_import_fiftyone
from lours.utils.difftools import dataset_diff

if TYPE_CHECKING:
    import fiftyone as fo
else:
    fo = try_import_fiftyone()


class DatasetUpdateHandler(FileSystemEventHandler):
    """Class to update the dataset in fiftyone each time a change is detected"""

    def __init__(
        self,
        dataset: Dataset,
        fo_dataset: fo.Dataset,
        images_root: Path,
        annotations_root: Path,
        splits_to_read: list[str],
        event_queue: EventQueue,
        cooldown_time: float = 1,
    ) -> None:
        """Constructor of DatasetUpdateHandler

        The handler will wait for a cooldown period of time and then will cancel all
        changes detected by watchdog in the mean time. This allows to only run the
        dataset update once when multiple files are modified at the same time.

        Args:
            dataset: dataset object to compare the newly loaded caipy dataset with,
                should have already been converted to fiftyone
            fo_dataset: fiftyone dataset object corresponding to the lours dataset.
                will be used to remove modified or deleted data from the caipy folder.
            images_root: parameter given to caipy loading function, should be the same
                as the one used when constructing the already existing dataset
            annotations_root: parameter given to caipy loading function, should be the
                same as the one used when constructing the already existing dataset
            splits_to_read: parameter given to caipy loading function, should be the
                same as the one used when constructing the already existing dataset
            event_queue: queue object to know if other changes have been detected after
                the cooldown
            cooldown_time: time in seconds between a detected change and a dataset
                update. Every change event between first event and the cooldown will be
                ignored. Useful when adding multiple files at once. Defaults to 1
        """
        super().__init__()
        self.dataset = dataset
        self.fo_dataset = fo_dataset
        self.images_root = images_root
        self.annotations_root = annotations_root
        self.splits_to_read = splits_to_read
        self.cooldown_time = cooldown_time
        self.event_queue = event_queue

    def update_dataset(self) -> None:
        """Compare the dataset with the one loaded from caipy.

        If different, add the new elements to fiftyone, but only them.
        """
        caipy_dataset = from_caipy_generic(
            images_folder=self.images_root,
            annotations_folder=self.annotations_root,
            splits_to_read=self.splits_to_read,
        )
        to_update, to_remove, common = dataset_diff(caipy_dataset, self.dataset)
        if len(to_remove) > 0:
            annotations_filter = fo.ViewField("lours_id").is_in(
                common.annotations.index
            )
            images_filter = fo.ViewField("lours_id").is_in(common.images.index)

            filtered_view = self.fo_dataset.match(images_filter)
            if "groundtruth_detection" in self.fo_dataset.get_field_schema():
                filtered_view = filtered_view.filter_labels(
                    "groundtruth_detection.detections", annotations_filter
                )
            if "groundtruth_keypoints" in self.fo_dataset.get_field_schema():
                filtered_view = filtered_view.filter_labels(
                    "groundtruth_keypoint.keypoints", annotations_filter
                )
            # Register this view on top of the existing dataset
            filtered_view.save()
            filtered_view.keep()
            filtered_view.keep_fields()

        if len(to_update) > 0:
            self.fo_dataset = to_update.to_fiftyone(
                dataset_name=str(self.fo_dataset.name), existing="update"
            )
        self.fo_dataset.save()
        self.dataset = caipy_dataset

    def on_any_event(self, event: FileSystemEvent) -> None:
        """Method called when a file or folder is created, modified or deleted.

        If after a cooldown poeriod, the event queue has elements, this will remove
        them from the queue so that the dataset update is not called again. Since the
        dataset update reads the whole folder, the subsequent detected changes will not
        be ignored even though they are removed from the event queue.

        Args:
            event: event describing the file or folder that was created.
                Not used today, but could probably be used in the future to avoid
                unnecessary reloading of caipy dataset
        """
        if event.event_type in [EVENT_TYPE_CLOSED, EVENT_TYPE_OPENED]:
            return
        sleep(self.cooldown_time)
        while not self.event_queue.empty():
            self.event_queue.get()
        self.update_dataset()


def get_argparser() -> ArgumentParser:
    """Function to get the argparser, which will be parsed itself by sphinx-argparse"""
    parser = ArgumentParser(
        description=(
            "Directly convert a Caipy folder into fiftyone for quick inspection"
        ),
    )

    i_parser = parser.add_argument_group("Input arguments")
    i_parser.add_argument(
        "--mode",
        choices=["vanilla", "generic"],
        default="vanilla",
        help=(
            "Choice between vanilla mode, where a single folder is given with option"
            " ``--input_folder``, and generic, where two folders are given with options"
            " ``--images_root`` and ``--annotations_root``"
        ),
    )
    i_parser.add_argument(
        "--input-folder",
        "-i",
        help="When in vanilla mode, folder where the CAIPY dataset is stored",
        type=Path,
        default=None,
    )
    i_parser.add_argument(
        "--images_root",
        "--ir",
        help=(
            "When in generic mode, folder where images are stored. Equivalent folder in"
            " vanilla mode is ``dataset/Images``"
        ),
        type=Path,
        default=None,
    )
    i_parser.add_argument(
        "--annotations_root",
        "--ar",
        help=(
            "When in generic mode, folder where annotations json files are stored."
            " Equivalent folder in vanilla mode is ``dataset/Annotations``"
        ),
        type=Path,
        default=None,
    )
    i_parser.add_argument(
        "--splits-to-read",
        "-s",
        nargs="*",
        default=None,
        help=(
            "Optional list of splits to read. If not selected, will read all splits and"
            " convert them to fiftyone"
        ),
    )
    f_parser = parser.add_argument_group("Fiftyone arguments")
    f_parser.add_argument(
        "--dataset-name",
        "--name",
        "-n",
        help=(
            "Optional dataset name to appear in fiftyone app. If not selected, will"
            " take the name of loaded dataset object, i.e. the name of ``input-folder``"
            " or parent of ``--images-root`` without its parents"
        ),
        default=None,
    )
    f_parser.add_argument(
        "--not-persistent",
        "--np",
        action="store_false",
        dest="persistent",
        help=(
            "If selected, will not save the converted dataset in fiftyone. It will be "
            "available to inspect with the browser during the time this script is"
            "running, but will be removed as soon as the script is stopped."
        ),
    )
    f_parser.add_argument(
        "--no-app",
        "--na",
        action="store_false",
        dest="launch_app",
        help=(
            "If selected, will not launch the app. This needs the 'persistent' option"
            " to be selected for the command to do anything"
        ),
    )
    f_parser.add_argument(
        "--watch",
        "-w",
        action="store_true",
        help=(
            "If selected, will watch the folder containing the dataset. Each time a"
            " change in the folder is detected, the dataset is updated and so is the"
            " fiftyone counterpart. Useful when you are constructing a dataset"
            " gradually. Note that this option is useless if the app is not launched"
        ),
    )
    f_parser.add_argument(
        "--port",
        "-p",
        default=0,
        type=int,
        help=(
            "Server port to connect to the app server. If not set, will choose randomly"
            " a free port"
        ),
    )
    f_parser.add_argument(
        "--only-local",
        action="store_true",
        help=(
            "if selected, will not open the server for other than localhost. the server"
            " can still be reachable with ssh tunnelling though"
        ),
    )
    return parser


def run():
    """CLI function for caipy to fiftyone"""
    parser = get_argparser()

    args = parser.parse_args()

    if not args.launch_app and not args.persistent:
        warn(
            "App won't be launched and dataset is not persistent, this command will not"
            " do anything",
            RuntimeWarning,
        )

    if args.mode == "vanilla":
        assert (
            args.input_folder is not None
        ), "You must provide a path to input_folder for vanilla mode"
        annotations_root = args.input_folder / "Annotations"
        images_root = args.input_folder / "Images"
        dataset = from_caipy(args.input_folder, splits_to_read=args.splits_to_read)
    else:
        assert args.images_root is not None and args.annotations_root is not None, (
            "You must provide paths for both images_root and annotations_root for"
            " generic mode"
        )
        images_root = args.images_root
        annotations_root = args.annotations_root
        dataset = from_caipy_generic(
            images_folder=args.images_root,
            annotations_folder=args.annotations_root,
            splits_to_read=args.splits_to_read,
        )

    dataset_name = (
        args.dataset_name if args.dataset_name is not None else dataset.dataset_name
    )

    fo_dataset = dataset.to_fiftyone(dataset_name=dataset_name)
    if args.persistent:
        fo_dataset.persistent = True

    if args.launch_app:
        launch_kwargs = {}
        if not args.only_local:
            launch_kwargs["address"] = "0.0.0.0"
        if args.port == 0:
            print("Getting a random free port ...")
            # Stolen from https://stackoverflow.com/a/61685162
            with socketserver.TCPServer(("localhost", 0), None) as s:  # pyright: ignore
                port = s.server_address[1]
            print(f"Port chosen : {port}")
        else:
            port = args.port
        launch_kwargs["port"] = port
        session = fo.launch_app(dataset=fo_dataset, **launch_kwargs)
        if args.watch:
            observer = Observer()
            event_handler = DatasetUpdateHandler(
                dataset,
                fo_dataset,
                images_root,
                annotations_root,
                args.splits_to_read,
                observer.event_queue,
            )
            observer.schedule(event_handler, str(annotations_root), recursive=True)
            observer.start()
        session.wait(-1)



import json
import shutil
from collections.abc import Iterable, Mapping
from os.path import relpath
from pathlib import Path
from typing import Any

import pandas as pd
import yaml
from imageio.v3 import imread, imwrite
from tqdm.auto import tqdm

from lours.utils.bbox_converter import column_names_from_format_string, export_bbox
from lours.utils.testing import assert_images_valid

from ..dataset import Dataset
from .common import (
    convert_str,
    get_image_info,
    get_images_from_folder,
    get_relative_image_path,
    to_dataset_object,
)

CX_COLUMN, W_COLUMN, CY_COLUMN, H_COLUMN = column_names_from_format_string("cxwcyh")


def bbox_to_txt(bbox_df: pd.DataFrame) -> list[str]:
    """Convert dataframe of yolo bboxes to a list of strings to be written in
    a file readable by darknet

    Args:
        bbox_df: bboxes dataframe. Must be in yolo format (see function above)

    Returns:
        list of string describing bboxes according to darknet format
    """
    lines = []
    for _, box in bbox_df.iterrows():
        box_string = (
            f"{box['category_id']} "
            f"{box[CX_COLUMN]} "
            f"{box[CY_COLUMN]} "
            f"{box[W_COLUMN]} "
            f"{box[H_COLUMN]}"
        )
        lines.append(f"{box_string}\n")
    return lines


def txt_to_bbox(text_lines: Iterable[str], image_id: int) -> list[dict[str, Any]]:
    """Read lines coming from a darknet txt annotation file and convert it
    to a list of annotation dictionaries

    Args:
        text_lines: list of text lines in a particular folder all given lines are
            assumed to come from the same image
        image_id: image index of corresponding image

    Returns:
        list of annotations for the particular image. bboxes are in the darknet format,
        i.e. with relative coordinates
    """
    bboxes = []
    for line in text_lines:
        cat_id, x, y, width, height, *rest = map(convert_str, line.split())
        bbox = {
            "category_id": cat_id,
            CX_COLUMN: x,
            CY_COLUMN: y,
            W_COLUMN: width,
            H_COLUMN: height,
            "image_id": image_id,
        }
        if rest:
            bbox["confidence"] = rest[0]
        bboxes.append(bbox)
    return bboxes


def open_data_file(data_file: Path) -> dict[str, str | int | float]:
    """Open a ``.data`` file typically used in darknet.

    Args:
        data_file: path to ``.data`` file

    Returns:
        dictionary of values contained in the .data file
    """
    data_dict = {}
    with open(data_file) as f:
        lines = f.read().split("\n")
    for line in lines:
        if line:
            try:
                key, value = line.split(" = ")
            except ValueError:
                key = line.strip("=").strip()
                value = ""
            data_dict[key] = convert_str(value)
    return data_dict


def write_data_file(data: dict[str, str | int], data_file: Path) -> None:
    """Save a dictionary to a ``.data`` file typically used in darknet.
    Note that only the typical keys expected by darknet are considered,
    i.e. classes, train, valid, names and backup

    Args:
        data: content to save to the ``.data`` file
        data_file: path where to save the given data dictionary
    """
    keys = ["classes", "train", "valid", "names", "backup"]
    with open(data_file, "w") as f:
        for key in keys:
            f.write(f"{key} = {data.get(key, '')}\n")


def yolov5_img_path_to_label_path(img_path: Path) -> Path:
    """Convert the path of an image to its corresponding label

    That is, if the path has a folder named "images", rename that folder into "labels"
    and change the image extension to "txt"

    See `Yolov5's image2labels <https://github.com/ultralytics/yolov5/blob/master/utils/dataloaders.py#L424>`__ .

    Args:
        img_path: path to an image file

    Returns:
        corresponding label file as it would have been searched for by yolov5
    """  # noqa: E501
    annotation_path = img_path.with_suffix(".txt")
    folder_parts = list(annotation_path.parts)
    if "images" in folder_parts:
        # In case they are multiple occurrences of images folder, use the last one
        images_folder_index = len(folder_parts) - 1 - folder_parts[::-1].index("images")
        folder_parts[images_folder_index] = "labels"
        annotation_path = Path(*folder_parts)
    return annotation_path


def deduce_data_file_and_dataset_folder(
    dataset_path: Path | str | None = None,
    data_file: Path | str | None = None,
    default_data_file: str = "train_job.data",
) -> tuple[Path, Path]:
    """Get dataset path and data file, deduce one from another if needed."""
    if dataset_path is None and data_file is None:
        raise ValueError("dataset_path and data_file cannot be both None !")
    if dataset_path is not None:
        dataset_path = Path(dataset_path)
        if not dataset_path.is_dir():
            raise ValueError("Dataset path must be a path to valid directory")
        if data_file is None:
            data_file = dataset_path / default_data_file
            if not data_file.is_file():
                raise ValueError(f"Data file {data_file} does not exist")
    if data_file is not None:
        data_file = Path(data_file)
        if not data_file.is_file():
            raise ValueError(f"Data file {data_file} does not exist")
        if dataset_path is None:
            dataset_path = data_file.parent

    # Do a final assert, otherwise pyright will complain
    assert dataset_path is not None and data_file is not None
    return dataset_path, data_file


def from_darknet(
    dataset_path: Path | str | None = None,
    data_file: Path | str | None = None,
    ids_map: dict[int, dict[str, Any]] | str | Path | None = None,
    image_info: pd.DataFrame | None = None,
) -> Dataset:
    """Creates dataset object from a darknet dataset. Note that category ids and
    image ids are not given in the dataset format and thus can only be sequential
    As such, if we want to convert the dataset back to another format that keeps track
    of image and category ids, we need to give image_info and class mapping from
    an external source. Here we expect it to be contained in a json label map for
    annotations and a DataFrame with similar columns as in the final dataset's images
    DataFrame for images

    Args:
        dataset_path: folder containing the dataset, from which the relative path
            are given. If not set, will use ``data_file``'s parent directory.
            Defaults to None
        data_file: data file containing info about names, lists of train and
            validation images. Can be either a .data file or a .yml file (for yolov5).
            If not set, will use the file ``train_job.data`` at the root of
            ``dataset_path``. Defaults to None.
        ids_map: Optional dictionary containing the id_remapping that was initially
            applied to create the darknet dataset. Will reverse it to get back to the
            original class mapping. The dictionary must have darknet dataset's category
            ids (in sequential order then) as keys and with corresponding values that
            are dictionaries containing ``name`` and ``id`` keys relative to this
            class. Note that this can also be a path to a json file containing the
            dictionary. Defaults to None.
        image_info: Optional DataFrame containing image information. Must contain at
            least the following columns : ``relative_path``, ``id``, ``width``,
            ``height``. Defaults to None

    Raises:
        ValueError: Errors when neither ``dataset_path`` nor ``data_file`` is specified

    Returns:
        Loaded dataset object
    """
    dataset_path, data_file = deduce_data_file_and_dataset_folder(
        dataset_path, data_file
    )

    if data_file.suffix in [".yml", ".yaml"]:
        return from_darknet_yolov5(
            dataset_path, data_file, ids_map=ids_map, image_info=image_info
        )

    if isinstance(ids_map, Path | str):
        with open(ids_map) as f:
            ids_map = json.load(f)
        assert isinstance(ids_map, dict)

    data_dict = open_data_file(data_file)
    train_file = data_dict["train"]
    valid_file = data_dict["valid"]
    names_file = data_dict["names"]
    assert isinstance(train_file, str)
    assert isinstance(valid_file, str)
    assert isinstance(names_file, str)
    # Filter is here to dismiss blank lines, especially at the end of the file
    with open(dataset_path / names_file) as f:
        name_list = filter(bool, f.read().split("\n"))
        dict(enumerate(name_list))

    if train_file:
        with open(dataset_path / train_file) as f:
            train_list = filter(bool, f.read().split("\n"))
    else:
        train_list = []
    with open(dataset_path / valid_file) as f:
        valid_list = filter(bool, f.read().split("\n"))

    dataset = from_darknet_generic(
        dataset_path,
        dataset_path,
        name_list,
        valid_list,
        "valid",
        ids_map=ids_map,
        image_info=image_info,
    )
    if train_list:
        dataset += from_darknet_generic(
            dataset_path,
            dataset_path,
            name_list,
            train_list,
            "train",
            ids_map=ids_map,
            image_info=image_info,
        )
    return dataset


def from_darknet_yolov5(
    dataset_path: Path | str | None = None,
    data_file: Path | str | None = None,
    splits: Iterable[str] | None = None,
    split_name_mapping: Mapping[str, str] | None = None,
    ids_map: dict[int, dict[str, Any]] | str | Path | None = None,
    image_info: pd.DataFrame | None = None,
) -> Dataset:
    """Creates dataset object from a darknet dataset. Note that category ids and
    image ids are not given in the dataset format and thus can only be sequential
    As such, if we want to convert the dataset back to another format that keeps track
    of image and category ids, we need to give image_info and class mapping from
    an external source. Here we expect it to be contained in a json label map for
    annotations and a DataFrame with similar columns as in the final dataset's images
    DataFrame for images

    Args:
        dataset_path: folder containing the dataset, from which the relative path
            are given. If not set, will use ``data_file``'s parent directory.
            Defaults to None
        data_file: data file containing info about names, lists of train and
            validation images. Must be a valid path to a Yaml file
            (either .yml or .yaml). If set to None, will use the file
            ``data.yaml`` at the root of ``dataset_path``. Defaults to None.
        splits: name of splits to load. if set to None, will consider every key in
            ``data_file`` that is neither ``names`` nor ``path`` to be a split.
            Every value in said keys must be either a valid folder or a valid text file,
            both relative to ``dataset_path``. Defaults to None.
        split_name_mapping: mapping dict to replace split names to other ones. split
            names not present in mapping will not be modified. If set to None,
            will apply yolov5 conventional mapping, i.e. 'val' -> 'valid' and
            'test' -> 'eval'. Defaults to None
        ids_map: Optional dictionary containing the id_remapping that was initially
            applied to create the darknet dataset. Will reverse it to get back to the
            original class mapping. The dictionary must have darknet dataset's category
            ids (in sequential order then) as keys and with corresponding values that
            are dictionaries containing ``name`` and ``id`` keys relative to this
            class. Note that this can also be a path to a json file containing the
            dictionary. Defaults to None.
        image_info: Optional DataFrame containing image information. Must contain at
            least the following columns : ``relative_path``, ``id``, ``width``,
            ``height``. Defaults to None

    Raises:
        ValueError: Errors when neither ``dataset_path`` nor ``data_file`` is specified

    Returns:
        Loaded dataset object
    """

    def get_folders_and_file(
        root_path: Path, data_path: str
    ) -> tuple[Path, list[str] | None]:
        full_path = root_path / data_path
        if full_path.is_dir():
            return full_path, None
        elif full_path.is_file():
            with open(full_path) as f:
                file_list = list(filter(bool, f.read().split("\n")))
            return root_path, file_list
        else:
            raise ValueError(f"Error, {full_path} is neither a folder nor a text file")

    if split_name_mapping is None:
        split_name_mapping = {"val": "valid", "test": "eval"}
    dataset_path, data_file = deduce_data_file_and_dataset_folder(
        dataset_path, data_file, default_data_file="data.yaml"
    )
    assert data_file.suffix in [".yml", ".yaml"]

    with open(data_file) as f:
        data_dict = yaml.safe_load(f)
    dataset_root = Path(data_dict.get("path", ""))
    if not dataset_root.is_absolute():
        dataset_root = dataset_path / dataset_root

    split_datasets = []
    if splits is None:
        splits = [
            name
            for name, value in data_dict.items()
            if name not in ["path", "names"] and isinstance(value, str)
        ]

    if not splits:
        raise ValueError("Not split was found or specified")

    for split_name in splits:
        split_path = data_dict[split_name]
        split_folder, split_files = get_folders_and_file(dataset_root, split_path)
        split_dataset = from_darknet_generic(
            split_folder,
            split_folder,
            data_dict["names"],
            split_files,
            split_name_mapping.get(split_name, split_name),
            ids_map,
            image_info,
        )
        split_datasets.append(split_dataset)
    return sum(split_datasets)  # pyright: ignore


def from_darknet_generic(
    images_root: Path | str,
    labels_root: Path | str,
    names: Iterable[str],
    image_files_list: Iterable[str | Path] | None = None,
    split: str | None = None,
    ids_map: dict[int, dict[str, Any]] | str | Path | None = None,
    image_info: pd.DataFrame | None = None,
) -> Dataset:
    """Generic function to load a darknet like dataset by only giving it folders,
    class names and optionally file list instead of a data file.

    Note:
        Unlike the darknet and yolov5 loaders, this function does only accept a single
        split value.

    Note:
        If no file list is given, the function will simply glob all image files in
        ``images_root`` folder.

    Args:
        images_root: Folder containing the image files to load
        labels_root: Folder containing the txt annotations files. Each annotation file
            is obtained by using the same relative path as the image equivalent and
            replacing the extension by txt. If no such file exist, it will not error
            but will assume there is no annotation
        names: list of classes used in the label map when saving this darknet dataset.
        image_files_list: list of images to read. Their path must be relative to
            ``images_root``. If set to None, will glob all files in ``images_root``
        split: Split name of the constructed dataset. Defaults to None.
        ids_map: Optional dictionary containing the id_remapping that was initially
            applied to create the darknet dataset. Will reverse it to get back to the
            original class mapping. The dictionary must have darknet dataset's category
            ids (in sequential order then) as keys and with corresponding values that
            are dictionaries containing ``name`` and ``id`` keys relative to this
            class. Note that this can also be a path to a json file containing the
            dictionary. Defaults to None.
        image_info: Optional DataFrame containing image information. Must contain at
            least the following columns : ``relative_path``, ``id``, ``width``,
            ``height``. Defaults to None

    Raises:
        FileNotFoundError: when the annotation file corresponding to a particular image
            file can not be found, be it using usual darknet convention or yolov5 one.

    Returns:
        Loaded dataset object
    """
    images_root = Path(images_root)
    labels_root = Path(labels_root)
    if image_files_list is None:
        image_files_list = get_images_from_folder(images_root)
    else:
        image_files_list = [Path(f) for f in image_files_list]
    images = []
    annotations = []

    if isinstance(ids_map, Path | str):
        with open(ids_map) as f:
            ids_map = json.load(f)
        assert isinstance(ids_map, dict)

    for image_number, image_path in enumerate(tqdm(image_files_list)):
        current_image = get_image_info(
            image_number=image_number,
            relative_path=image_path,
            absolute_path=images_root / image_path,
            image_info=image_info,
        )
        current_image["split"] = split
        images.append(current_image)

        annotation_path_simple = labels_root / image_path.with_suffix(".txt")
        annotation_path_yolov5 = labels_root / yolov5_img_path_to_label_path(image_path)

        if annotation_path_simple.is_file():
            annotation_path = annotation_path_simple
        elif annotation_path_yolov5.is_file():
            annotation_path = annotation_path_yolov5
        else:
            raise FileNotFoundError(
                f"Annotation could not be found. Neither {annotation_path_simple} nor"
                f" {annotation_path_yolov5} are files"
            )
        with open(annotation_path) as f:
            annotations_strings = filter(bool, f.read().split("\n"))
        annotations.extend(txt_to_bbox(annotations_strings, current_image["id"]))

    return to_dataset_object(
        images_root,
        dict(enumerate(names)),
        images,
        annotations,
        "cxwcyh",
        ids_map,
    )


def from_darknet_json(
    dataset_path: Path,
    json_path: Path,
    ids_map: dict[int, dict[str, Any]],
    image_info: pd.DataFrame | None,
    split_name: str = "eval",
) -> Dataset:
    """Same as from_darknet, expect the data file replaced with a json file containing
    directly annotations information. This is typically the format of predictions done
    by darknet's detector.

    Args:
        dataset_path: folder containing the dataset, from which
            the relative path are given
        json_path: json file containing a list of predictions as dictionaries. Each
            dictionary will have bbox info as well ad image path, which will be used
            to retrieve the original image id thanks to image_info DataFrame
        ids_map: dictionary containing the id_remapping that was initially applied to
            create the darknet dataset. Will reverse it to get back to the original
            class mapping. The dictionary must have darknet dataset's category ids
            (in sequential order then) as keys and with corresponding values that are
            dictionaries containing ``name`` and ``id`` keys relative to this class.
        image_info: DataFrame containing image information. Must contain at least the
            following columns : ``relative_path``, ``id``, ``width``, ``height``
        split_name: Name of the split that will be assigned to the ``split`` column of
            the resulting dataset's annotation dataframe.

    Returns:
        Loaded dataset object
    """
    images = []
    annotations = []
    with open(json_path) as f:
        json_list = json.load(f)
    for image_number, frame in enumerate(json_list):
        current_image = {}
        image_path = Path(frame["filename"])
        relative_path = get_relative_image_path(
            dataset_path=dataset_path, image_path=image_path
        )
        current_image["relative_path"] = relative_path
        current_image["type"] = image_path.suffix
        current_image["split"] = split_name
        absolute_path = dataset_path / relative_path
        current_image.update(
            **get_image_info(
                image_number=image_number,
                relative_path=relative_path,
                absolute_path=absolute_path,
                image_info=image_info,
            )
        )

        bboxes = []
        for current_object in frame["objects"]:
            coords = current_object["relative_coordinates"]
            bbox = {
                CX_COLUMN: coords["center_x"],
                CY_COLUMN: coords["center_y"],
                W_COLUMN: coords["width"],
                H_COLUMN: coords["height"],
                "category_id": current_object["class_id"],
                "category_str": current_object["name"],
                "confidence": current_object["confidence"],
                "image_id": current_image["id"],
            }
            bboxes.append(bbox)
        annotations.extend(bboxes)
        images.append(current_image)

    dataset = to_dataset_object(
        dataset_path,
        None,
        images,
        annotations,
        "cxwcyh",
        ids_map,
    )

    if image_info is not None:
        images_without_box = image_info[~image_info.index.isin(dataset.images.index)]
        dataset.images = pd.concat([image_info, images_without_box])
    dataset.images["split"] = split_name

    return dataset


def dataset_to_darknet(
    dataset: Dataset,
    output_path: Path | str,
    copy_images: bool = False,
    overwrite_images: bool = True,
    overwrite_labels: bool = True,
    yolo_version: int = 1,
    data_yaml_name: str = "data.yaml",
    split_name_mapping: dict[str, str] | None = None,
    create_split_folder: bool = False,
) -> None:
    """Save given dataset to darknet. Will recreate folder structure
    of dataset_path + image relative path. Will also create metadata files needed
    by darknet, i.e. names file, json label maps, list of train, validation and
    evaluation images (depending on available splits) and train and eval data files

    Args:
        dataset: dataset object to save as darknet
        output_path: root folder where images, corresponding text files
            describing bboxes will be saved along metadata files described above
        copy_images: If set to False, will create a symbolic link instead of copying.
            Much faster, but needs to keep original images in the same relative path.
            Defaults to False.
        overwrite_images: if set to False, will skip images that are already copied.
            Defaults to True.
        overwrite_labels: if set to False, will skip annotation that are already
            created. Defaults to True.
        yolo_version: version number for yolo. Officially supported versions are
            ``1`` to ``4`` (regular darknet), ``5`` (Ultralytics), ``7``
            (Chien-Yao Wang). Versions ``6`` and ``8`` are NOT officially supported, and
            will be assumed to all be version ``5``.
        data_yaml_name: if ``yolo_version`` is > 4, will be used as filename to save
            the yaml file at the root of ``output_path``
        split_name_mapping: mapping dict to replace split names to other ones. split
            names not present in mapping will not be modified. If set to None,
            will apply yolov5 conventional mapping,
            i.e. ``{'valid': 'val', 'validation': 'val', 'eval': 'test'}``.
            Defaults to None
        create_split_folder: if set to True, will create a dedicated folder for each
            split and will save images in it. Image paths in {split}.txt will be changed
            accordingly. Note that this changes the dataset structure. Defaults to False
    """
    try:
        assert_images_valid(dataset, load_images=False)
    except AssertionError as e:
        raise ValueError(
            "Dataset images are missing, check that the images root folder is the"
            " right one"
        ) from e
    if split_name_mapping is None:
        split_name_mapping = {"valid": "val", "validation": "val", "eval": "test"}

    output_path = Path(output_path)
    output_path.mkdir(exist_ok=True, parents=True)
    id_remapping = {
        i: {"id": cat_id, "name": cat_name}
        for i, (cat_id, cat_name) in enumerate(dataset.label_map.items())
    }

    # Remap classes so that category ids are now sequential
    sorted_dataset = dataset.remap_classes(
        {value["id"]: i for i, value in id_remapping.items()}
    )
    name_list = [*sorted_dataset.label_map.values()]
    name_list_path = output_path / "classes.names"
    with open(output_path / "classes.names", "w") as f:
        f.writelines(f"{name}\n" for name in name_list)
    with open(output_path / "ids_map.json", "w") as f:
        json.dump(id_remapping, f, indent=2)
    with open(output_path / "label_map.json", "w") as f:
        json.dump(dataset.label_map, f, indent=2)

    valid_splits = ["train", "valid", "eval"]

    if len(dataset.images["split"].isin(valid_splits)) < len(dataset.images):
        print(
            f"Darknet saver only takes these splits : {', '.join(valid_splits)}. "
            "Other values will be ignored"
        )

    remapped_splits = dataset.images["split"].replace(split_name_mapping)
    if create_split_folder:
        output_relative_path = pd.concat(
            [dataset.images["relative_path"], remapped_splits], axis=1
        ).apply(lambda x: x["split"] / x["relative_path"].with_suffix(".jpg"), axis=1)
    else:
        output_relative_path = dataset.images["relative_path"].apply(
            lambda x: x.with_suffix(".jpg")
        )
    split_path_lists = {
        split_name: path_list
        for split_name, path_list in output_relative_path.groupby(remapped_splits)
    }

    yaml_data = {
        "nc": len(name_list),
        "names": name_list,
    }
    for split_name, split_paths in split_path_lists.items():
        list_name = f"{split_name}.txt"
        list_path = output_path / list_name
        if yolo_version == 7:
            # For yolov7, paths need to be absolute, for both the txt file and the
            # image paths inside of it.
            # If possible, avoid saving to yolov7 format because it's hard to move
            # elsewhere afterwards

            # For the following pyright ignore, see the bug issue here
            # https://github.com/pandas-dev/pandas-stubs/issues/706
            # TODO : remove it when it is solved
            yaml_data[split_name] = str(list_path.absolute())  # pyright: ignore
            with open(list_path, "w") as f:
                f.writelines(
                    f"{output_path.absolute() / sample}\n" for sample in split_paths
                )
        else:
            yaml_data[split_name] = list_name  # pyright: ignore
            with open(list_path, "w") as f:
                f.writelines(f"{sample}\n" for sample in split_paths)

    if yolo_version <= 4:
        if len(split_path_lists.get("train", [])) > 0:  # pyright: ignore
            assert len(split_path_lists.get("val", [])) > 0  # pyright: ignore
        train_list_name = yaml_data["train"]
        val_list_name = yaml_data["val"]
        train_data_path = output_path / "train.data"
        write_data_file(
            {
                "classes": len(name_list),
                "train": train_list_name,
                "valid": val_list_name,
                "names": name_list_path.name,
            },
            train_data_path,
        )

        if len(split_path_lists.get("eval", [])) > 0:  # pyright: ignore
            eval_list_name = yaml_data["eval"]
            eval_data_path = output_path / "eval.data"
            write_data_file(
                {
                    "classes": len(name_list),
                    "valid": eval_list_name,
                    "names": name_list_path.name,
                },
                eval_data_path,
            )
    else:
        with open(output_path / data_yaml_name, "w") as f:
            yaml.safe_dump(yaml_data, f)

    bboxes_yolo = export_bbox(
        sorted_dataset.annotations, sorted_dataset.images, output_format="cxwcyh"
    )
    yolo_annotations = pd.concat([sorted_dataset.annotations, bboxes_yolo], axis=1)
    output_resolved = output_path.resolve()
    for image_id, image_data in tqdm(dataset.images.iterrows(), total=len(dataset)):
        instances = yolo_annotations[yolo_annotations["image_id"] == image_id]
        output_image_relative_path = output_relative_path.loc[
            image_id
        ]  # pyright: ignore
        input_image_path = (dataset.images_root / image_data["relative_path"]).resolve()
        output_image_path = output_resolved / output_image_relative_path
        output_image_path.parent.mkdir(exist_ok=True, parents=True)
        if yolo_version > 4:
            output_txt_path = yolov5_img_path_to_label_path(output_image_path)
        else:
            output_txt_path = output_image_path.with_suffix(".txt")
        if not output_image_path.is_file() or overwrite_images:
            if image_data["type"].lower() not in [".jpg", ".jpeg"]:
                image = imread(input_image_path)
                imwrite(output_image_path, image)
            elif copy_images:
                shutil.copy(input_image_path, output_image_path)
            else:
                output_image_path.unlink(missing_ok=True)
                output_image_path.symlink_to(
                    relpath(input_image_path, output_image_path.parent)
                )
        if not output_txt_path.is_file() or overwrite_labels:
            output_txt_path.parent.mkdir(parents=True, exist_ok=True)
            with open(output_txt_path, "w") as f:
                f.writelines(bbox_to_txt(instances))



from collections import defaultdict
from collections.abc import Sequence
from typing import Any, TypeVar

import numpy as np
import pandas as pd
from numpy import ndarray

from ..dataset import Dataset
from . import BBOX_COLUMN_NAMES
from .bbox_converter import import_bbox
from .label_map_merger import merge_label_maps

*_, WIDTH_NAME, HEIGHT_NAME = BBOX_COLUMN_NAMES

D = TypeVar("D", bound=Dataset)
Scalar = float | str | bool
ScalarSequence = Sequence[float] | Sequence[str] | Sequence[bool]


def broadcast_annotations(
    image_id: int | Sequence[int] | ndarray,
    bbox_coordinates: ndarray | Sequence[float] | Sequence[Sequence[float]],
    category_id: int | ndarray | Sequence[int],
    confidence: float | Sequence[float] | ndarray | None = None,
    **other: ndarray | Scalar | ScalarSequence,
) -> dict[str, np.ndarray]:
    """Broadcast together detection annotation attributes.

    Every attribute except ``bbox_coordinates`` can be only 1 element, and will be
    duplicated to match the length of ``bbox_coordinates``.

    If its more than 1 element, the length must match the length of ``bbox_coordinates``

    Args:
        image_id: id or 1D array of shapes [N] of ids of image corresponding to each
            annotation to be appended
        bbox_coordinates: 2D numpy array of shape [N, 4] or [N, 2] (for a keypoint)
            corresponding to the bounding box coordinates.
            Can also be of the shape [4] or [2] if there is only one
            bounding box/keypoint
        category_id: int value or 1D numpy array of shape [N] corresponding to
            the category of each bounding box
        confidence: when dealing with the special case of predictions, 1D numpy
            array of shape [N] corresponding to the confidence of the prediction.
            None otherwise. Defaults to None.
        **other: Other possible fields to the annotation

    Raises:
        ValueError: Will be raised if the bbox_coordinates shape is not [N, 4] [N, 2] or
            [2] or [4]
        ValueError: Will be raised if the values to broadcast are not scalar or array of
            size [1] or size [N]

    Returns:
        dictionary with the annotation attribute name as keys
        ("image_id", "category_id", etc) and the broadcast of values as numpy arrays
        as values.
    """
    if not isinstance(bbox_coordinates, ndarray):
        bbox_coordinates = np.array(bbox_coordinates)
    bbox_shape = bbox_coordinates.shape
    if len(bbox_shape) == 1:
        bbox_coordinates = bbox_coordinates[None]
    if (len(bbox_coordinates.shape) != 2) or (bbox_coordinates.shape[1] not in [2, 4]):
        raise ValueError(
            "Error for bbox_coordinates, expected shape of form [4], [2], [N, 4]"
            f" or [N, 2] got {bbox_shape}"
        )
    n_annot = bbox_coordinates.shape[0]

    def broadcast_scalar(
        value_to_broadcast: Scalar | ScalarSequence | np.ndarray,
        key_name: str,
    ) -> np.ndarray:
        if np.ndim(value_to_broadcast) == 0:
            broadcasted = np.tile(value_to_broadcast, n_annot)
        elif np.ndim(value_to_broadcast) == 1:
            value_to_broadcast_np = np.array(value_to_broadcast)
            if value_to_broadcast_np.shape[0] == 1:
                broadcasted = np.tile(value_to_broadcast_np, n_annot)
            elif value_to_broadcast_np.shape[0] == n_annot:
                broadcasted = value_to_broadcast_np
            else:
                broadcasted = None
        else:
            broadcasted = None
        if broadcasted is None:
            raise ValueError(
                f"Value to broadcast {key_name} can only be either a scalr of a 1D"
                f" dimensional array of length 1 or {n_annot}"
            )
        return broadcasted

    result = {}
    result["image_id"] = broadcast_scalar(image_id, "image_id")
    result["bbox_coordinates"] = bbox_coordinates
    result["category_id"] = broadcast_scalar(category_id, "category_id")
    if confidence is not None:
        if n_annot == 1:
            result["confidence"] = broadcast_scalar(confidence, "confidence")
        else:
            confidence = np.array(confidence)
            if confidence.shape != (n_annot,):
                raise ValueError(
                    "Confidence field can only be the same length as bounding boxes."
                    f" Expected length of {n_annot}, got shape of {confidence.shape}"
                )
            result["confidence"] = confidence

    for key, value in other.items():
        result[key] = broadcast_scalar(value, key)

    return result


def add_detection_annotation(
    input_dataset: D,
    image_id: int | Sequence[int] | ndarray,
    bbox_coordinates: Sequence[float] | Sequence[Sequence[float]] | ndarray,
    format_string: str,
    category_id: ndarray | int | Sequence[int],
    inplace: bool = False,
    label_map: dict[int, str] | None = None,
    category_ids_mapping: dict[int, int] | None = None,
    confidence: float | ndarray | Sequence[float] | None = None,
    **other_columns: Scalar | ScalarSequence | ndarray,
) -> D:
    """Add one or multiple detection annotations to the current dataset.
    In the case of a single annotation, every option can be a single value, but in the
    case of multiple annotations, every option needs to be an array of such values, and
    every array needs to be the same length.

    Note:
        In additions to the following options, you can add other fields as well, with
        keyword arguments.


    Args:
        input_dataset: Dataset to which we want to add new annotations
        image_id: image identifier to link each detection to the corresponding image
        bbox_coordinates: list of coordinates for the bounding box. Can follow any
            compatible format, as long as it is given in the next format
        format_string: format of coordinates, whether coordinates are relatives, using
            corner points of the box, box dimensions, etc. See :func:`.import_bbox` for
            more info
        category_id: category of each detection. Label will be deduced from dataset's
            label map
        inplace: if set to True, will modify the dataset in place and return itself.
            Else, will return a modified Dataset. Defaults to False.
        label_map: In the case the current dataset's label map is incomplete, merge it
            with this new label map. current label map and new label map must be
            compatible, see :func:`.merge_label_maps`. Defaults to None.
        category_ids_mapping: Optional dictionary to map annotated category ids into the
            right ids. This is useful for example when a neural network can only use a
            contiguous label map.
        confidence: Optional field for confidence, in the case annotations are actually
            predictions. Must the same length as bbox_coordinates. In the case of a
            single prediction, can also be a float. Defaults to None.
        **other_columns: Other column representing custom fields.

    Raises:
        ValueError: raised when giving numpy arrays are not the same number of elements,
            or if the bounding box coordinates is not of the shape either [4], or [N, 4]

    Returns:
        If ``inplace`` is False, new dataset object with appended annotations.
        Otherwise, the updated ``input_dataset``.
    """
    if label_map is not None:
        label_map = merge_label_maps(input_dataset.label_map, label_map, method="outer")
    else:
        label_map = input_dataset.label_map

    columns = broadcast_annotations(
        image_id=image_id,
        bbox_coordinates=bbox_coordinates,
        category_id=category_id,
        confidence=confidence,
        **other_columns,
    )
    bbox_coordinates = columns.pop("bbox_coordinates")
    for name, column in columns.items():
        if column.ndim != 1:
            raise ValueError(
                "Field as arrays can only be 1-dimensional, but got a shape of"
                f" {column.shape} for column {name}"
            )

    if bbox_coordinates.shape[1] not in [2, 4]:
        raise ValueError(
            "Error for bbox_coordinates, expected an array of 4 or 2 columns, got"
            f" shape of {bbox_coordinates.shape}"
        )
    annotations_df = pd.DataFrame(columns)
    if input_dataset.len_annot() > 0:
        first_index = input_dataset.annotations.index.max() + 1
        annotations_df.index += first_index
    if category_ids_mapping is not None:
        annotations_df["category_id"] = annotations_df["category_id"].replace(
            category_ids_mapping
        )

    bbox_df = import_bbox(
        bounding_boxes=bbox_coordinates,
        images_df=input_dataset.images,
        image_ids=annotations_df["image_id"],
        input_format=format_string,
    )

    bbox_df.index = annotations_df.index

    annotations_df = pd.concat([annotations_df, bbox_df], axis=1)

    if ("area" in input_dataset.annotations.columns) and (
        "area" not in annotations_df.columns
    ):
        annotations_df["area"] = bbox_df[WIDTH_NAME] * bbox_df[HEIGHT_NAME]

    annotations_df = pd.concat([input_dataset.annotations, annotations_df])

    if inplace:
        input_dataset.label_map = label_map
        input_dataset.annotations = annotations_df
        input_dataset.init_annotations()
        return input_dataset
    else:
        return input_dataset.from_template(
            annotations=annotations_df, label_map=label_map
        )


class AnnotationAppender:
    """Context manager to easily add detection tensors, as if the Dataset was a list
    after the appending is finished, the appender construct big numpy arrays to
    concatenate to the dataset's annotations dataframe
    """

    def __init__(
        self,
        dataset: Dataset,
        format_string: str = "XYWH",
        category_ids_mapping: dict[int, int] | None = None,
        label_map: dict[int, str] | None = None,
    ):
        """Main constructor.

        Args:
            dataset: dataset the annotations will be appended to.
            format_string: String describing bounding box format. Defaults to "XYWH".
            category_ids_mapping: Optional dictionary to map annotated category ids into
                the right ids. This is useful for example when a neural network can only
                use a contiguous label map.
            label_map: Optional dictionary to provide ``category_id -> ``category_str``
                mapping for additional categories in appended annotations.
                Defaults to None.
        """
        self.dataset = dataset
        self.image_ids = set(dataset.images.index)
        self.format_string = format_string
        self.label_map = label_map
        self.categroy_ids_mapping = category_ids_mapping
        self.reset()

    def reset(self):
        """Creates an empty dictionary that will be fed new annotations."""
        self.annotations_to_append = defaultdict(dict)
        self.i = 0
        self.index = {}

    def __enter__(self):
        """Function called at the beginning of ``with`` context.

        Returns:
            self class, with the ``append``method.
        """
        self.reset()
        return self

    def append(
        self,
        image_id: int | Sequence[int] | ndarray,
        bbox_coordinates: ndarray | Sequence[float],
        category_id: int | Sequence[int] | ndarray,
        confidence: ndarray | float | Sequence[float] | None = None,
        **other: Scalar | ScalarSequence | ndarray,
    ) -> None:
        """Append annotations for a particular image id. Everything except image id
        is expected to be a numpy array. Note that in addition to the regular bounding
        boxes coordinates, category and confidence, you can add other fields as long as
        they are numpy array of the same length. If no column exist in the dataset's
        annotations dataframe, it will be created (setting the already existing
        annotations to NaN for this column)

        Args:
            image_id: id or 1D array of shapes [N] of ids of image corresponding to each
                annotation to be appended
            bbox_coordinates: 2D numpy array of shape [N, 4] corresponding to the
                bounding box coordinates
            category_id: hashable value or 1D numpy array of shape [N] corresponding to
                the category of each bounding box
            confidence: when dealing with the special case of predictions, 1D numpy
                array of shape [N] corresponding to the confidence of the prediction,
                which can also be a float if N == 1. None otherwise. Defaults to None.
            **other: Other possible fields to the annotation

        Raises:
            ValueError: raised when the given numpy arrays are not of the same length,
                or if bounding box coordinates are not a 2D array with 4 columns.
            ValueError: raised when the given image id is not present in the dataset's
                image dataframe
        """
        ids_to_check = [image_id] if isinstance(image_id, int) else image_id
        for i in ids_to_check:
            if i not in self.image_ids:
                raise ValueError(f"Image id {i} is not in dataset's images dataframe")
        broadcasted_annotations = broadcast_annotations(
            image_id=image_id,
            bbox_coordinates=bbox_coordinates,
            category_id=category_id,
            confidence=confidence,
            **other,
        )
        n_annot = len(broadcasted_annotations["image_id"])
        # Note that we need index because not all append calls have the same keywords
        # As such, the index can help up know which rows should have actual value,
        # and which rows should have pd.NA
        self.index[self.i] = np.arange(self.i, self.i + n_annot)

        for key, value in broadcasted_annotations.items():
            self.annotations_to_append[key][self.i] = value
        self.i += n_annot

    def finish(self) -> None:
        """Concatenate all annotations given by ``append`` method into one dataframe,
        with the right bounding box format and concatenate it to the Dataset's
        annotation original DataFrame.
        """
        concatenated_annotations = []
        try:
            bbox = np.concatenate(
                list(self.annotations_to_append.pop("bbox_coordinates").values())
            )
        except KeyError:
            # No bbox coordinates were added
            return
        for name, annot in self.annotations_to_append.items():
            index = np.concatenate([self.index[k] for k in annot.keys()])
            array = np.concatenate(list(annot.values()))
            concatenated_annotations.append(pd.Series(array, index=index, name=name))
        annotations_to_append = pd.concat(concatenated_annotations, axis=1)

        annotations_to_append_dict = {
            str(k): annotations_to_append[k].to_numpy()
            for k in annotations_to_append.columns
        }
        add_detection_annotation(
            self.dataset,
            format_string=self.format_string,
            inplace=True,
            label_map=self.label_map,
            category_ids_mapping=self.categroy_ids_mapping,
            bbox_coordinates=bbox,
            **annotations_to_append_dict,
        )

    def __exit__(self, exit_type: Any, exit_value: Any, traceback: Any) -> None:
        """Function called at the en of context, when annotations have been appended.

        Args:
            exit_type: For context manager compatibility. Not used here.
            exit_value: For context manager compatibility. Not used here.
            traceback: For context manager compatibility. Not used here.
        """
        self.finish()
